/**
 * Copyright (C) 2020-2022 Dušan Vejnović <vaadin@dussan.org>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.dussan.vaadin.dvirtual.config;

import java.io.FileReader;
import java.io.IOException;
import java.io.Serializable;
import java.nio.file.Paths;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.Objects;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;

import org.apache.commons.collections4.CollectionUtils;
import org.apache.commons.lang3.StringUtils;
import org.dussan.vaadin.dvirtual.database.data.ApplicationData;
import org.dussan.vaadin.dvirtual.database.data.ServersGroupData;
import org.dussan.vaadin.dvirtual.utils.LoggerUtil;
import org.slf4j.Logger;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;

public class ApplicationConfig
    implements Serializable
{

    private static final long serialVersionUID = -2997412210413696025L;

    private static final Logger LOGGER = LoggerUtil.getLogger();

    private static final String APPLICATION = "application";

    private static final String APPLICATION_COMMAND = "command";

    private static final String APPLICATION_ENABLED = "enabled";

    private static final String APPLICATION_ID = "id";

    private static final String APPLICATION_NAME = "name";

    private static final String APPLICATION_NOTE = "note";

    private static final String APPLICATION_ONE_INSTANCE = "one_instance";

    private static final String APPLICATION_SERVER_AUTO_SELECT = "server_auto_select";

    private static final String APPLICATION_SERVERS_GROUP_ID = "servers_group_id";

    private static final String APPLICATION_SERVERS_GROUP_IDS = "servers_group_ids";

    private static final String APPLICATION_SERVERS_GROUP_NAME = "servers_group_name";

    private static final String APPLICATION_SERVERS_GROUP_NAMES = "servers_group_names";

    private static final String DATA_VERSION = "version";

    private static final String DATA_VERSION_1 = "1";

    private static final String ERROR_CANNOT_LOAD = "Cannot load application configuration, cause: {}";

    private static final String SAX_FEATURES_EXTERNAL_GENERAL_ENTITIES = "http://xml.org/sax/features/external-general-entities";

    private static final String SPLITTER_COMMAS = ",";

    private static final String SPLITTER_SPACE = " ";

    private transient List<ApplicationData> applicationDatas = null;

    private ApplicationConfig()
    {
        applicationDatas = new ArrayList<>();
    }

    public ApplicationConfig( String configFolder, String configFile )
    {
        this();
        try (FileReader fileReader = new FileReader( Paths.get( configFolder, configFile )
                                                          .toFile() ))
        {
            DocumentBuilderFactory documentBuilderFactory = DocumentBuilderFactory.newDefaultInstance();
            documentBuilderFactory.setFeature( SAX_FEATURES_EXTERNAL_GENERAL_ENTITIES, false );
            documentBuilderFactory.setValidating( false );
            DocumentBuilder documentBuilder = documentBuilderFactory.newDocumentBuilder();
            Document document = documentBuilder.parse( new InputSource( fileReader ) );
            Element rootNode = document.getDocumentElement();

            // get data version
            String dataVersion = null;
            for ( int index = 0; index < rootNode.getAttributes()
                                                 .getLength(); index++ )
            {
                Node node = rootNode.getAttributes()
                                    .item( index );
                if ( Objects.nonNull( node ) && StringUtils.equals( node.getNodeName(), DATA_VERSION ) && StringUtils.isNotBlank( node.getTextContent() ) )
                {
                    dataVersion = StringUtils.stripToEmpty( node.getTextContent() );
                }
            }

            // application data
            if ( StringUtils.equals( dataVersion, DATA_VERSION_1 ) )
            {
                NodeList nodeList = rootNode.getElementsByTagName( APPLICATION );
                for ( int index = 0; index < nodeList.getLength(); index++ )
                {
                    Node applicationNode = nodeList.item( index );
                    if ( applicationNode.getNodeType() == Node.ELEMENT_NODE && applicationNode.hasChildNodes() )
                    {
                        ApplicationData applicationData = new ApplicationData();
                        applicationData.setId( ApplicationData.NEW_APPLICATION );
                        applicationData.setServersGroupDatas( new ArrayList<>() );

                        NodeList applicationChildNodeList = applicationNode.getChildNodes();
                        for ( int i = 0; i < applicationChildNodeList.getLength(); i++ )
                        {
                            Node applicationChildNode = applicationChildNodeList.item( i );
                            String applicationNodeText = StringUtils.stripToEmpty( applicationChildNode.getTextContent() );
                            switch ( applicationChildNode.getNodeName() )
                            {
                                case APPLICATION_COMMAND:
                                    applicationData.setCommand( applicationNodeText );
                                    break;
                                case APPLICATION_ENABLED:
                                    applicationData.setEnabled( Boolean.parseBoolean( applicationNodeText ) );
                                    break;
                                case APPLICATION_ONE_INSTANCE:
                                    if ( StringUtils.isNumeric( applicationNodeText ) )
                                    {
                                        applicationData.setOneInstance( Math.abs( Integer.parseInt( applicationNodeText ) ) );
                                    }
                                    break;
                                case APPLICATION_ID:
                                    if ( StringUtils.isNumeric( applicationNodeText ) )
                                    {
                                        applicationData.setId( Math.abs( Long.parseLong( applicationNodeText ) ) );
                                    }
                                    break;
                                case APPLICATION_NAME:
                                    applicationData.setName( applicationNodeText );
                                    break;
                                case APPLICATION_NOTE:
                                    applicationData.setNote( applicationNodeText );
                                    break;
                                case APPLICATION_SERVER_AUTO_SELECT:
                                    applicationData.setServerAutoSelect( Boolean.parseBoolean( applicationNodeText ) );
                                    break;
                                case APPLICATION_SERVERS_GROUP_ID:
                                    if ( StringUtils.isNumeric( applicationNodeText ) )
                                    {
                                        List<ServersGroupData> datas = applicationData.getServersGroupDatas();
                                        ServersGroupData serversGroupData = new ServersGroupData();
                                        serversGroupData.setId( Math.abs( Long.parseLong( applicationNodeText ) ) );
                                        datas.add( serversGroupData );
                                        applicationData.setServersGroupDatas( datas );
                                    }
                                    break;
                                case APPLICATION_SERVERS_GROUP_IDS:
                                    if ( StringUtils.isNumericSpace( StringUtils.replace( applicationNodeText, SPLITTER_COMMAS, SPLITTER_SPACE ) ) )
                                    {
                                        List<ServersGroupData> datas = applicationData.getServersGroupDatas();
                                        Arrays.asList( StringUtils.split( StringUtils.replace( applicationNodeText, SPLITTER_SPACE, SPLITTER_COMMAS ), SPLITTER_COMMAS ) )
                                              .stream()
                                              .filter( data -> StringUtils.isNotEmpty( StringUtils.stripToEmpty( data ) ) )
                                              .forEach( data -> {
                                                  ServersGroupData serversGroupData = new ServersGroupData();
                                                  serversGroupData.setId( Math.abs( Long.parseLong( StringUtils.strip( data ) ) ) );
                                                  datas.add( serversGroupData );
                                              } );
                                        applicationData.setServersGroupDatas( datas );
                                    }
                                    break;
                                case APPLICATION_SERVERS_GROUP_NAME:
                                    if ( StringUtils.isNotBlank( applicationNodeText ) )
                                    {
                                        List<ServersGroupData> datas = applicationData.getServersGroupDatas();
                                        ServersGroupData serversGroupData = new ServersGroupData();
                                        serversGroupData.setName( applicationNodeText );
                                        datas.add( serversGroupData );
                                        applicationData.setServersGroupDatas( datas );
                                    }
                                    break;
                                case APPLICATION_SERVERS_GROUP_NAMES:
                                    if ( StringUtils.isNotBlank( applicationNodeText ) )
                                    {
                                        List<ServersGroupData> datas = applicationData.getServersGroupDatas();
                                        Arrays.asList( StringUtils.split( applicationNodeText, SPLITTER_COMMAS ) )
                                              .stream()
                                              .filter( data -> StringUtils.isNotEmpty( StringUtils.stripToEmpty( data ) ) )
                                              .forEach( data -> {
                                                  ServersGroupData serversGroupData = new ServersGroupData();
                                                  serversGroupData.setName( StringUtils.strip( data ) );
                                                  datas.add( serversGroupData );
                                              } );
                                        applicationData.setServersGroupDatas( datas );
                                    }
                                    break;
                                default:
                                    // nothing to do
                                    break;
                            }
                        }

                        // add new application data
                        if ( StringUtils.isNotBlank( applicationData.getName() ) && CollectionUtils.isNotEmpty( applicationData.getServersGroupDatas() ) )
                        {
                            applicationDatas.add( applicationData );
                        }
                    }
                }
            }
        }
        catch ( IOException | SAXException | ParserConfigurationException exception )
        {
            LOGGER.info( ERROR_CANNOT_LOAD, exception.getMessage() );
        }
    }

    public List<ApplicationData> getApplicationDatas()
    {
        return applicationDatas;
    }

}
