/**
 * Copyright (C) 2020-2022 Dušan Vejnović <vaadin@dussan.org>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.dussan.vaadin.dvirtual.config;

import java.io.FileReader;
import java.io.IOException;
import java.io.Serializable;
import java.nio.file.Paths;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.Objects;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;

import org.apache.commons.collections4.CollectionUtils;
import org.apache.commons.lang3.StringUtils;
import org.dussan.vaadin.dvirtual.database.data.ApplicationData;
import org.dussan.vaadin.dvirtual.database.data.ApplicationsGroupData;
import org.dussan.vaadin.dvirtual.utils.LoggerUtil;
import org.slf4j.Logger;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;

public class ApplicationsGroupConfig
    implements Serializable
{

    private static final long serialVersionUID = 854891207425678489L;

    private static final Logger LOGGER = LoggerUtil.getLogger();

    private static final String APPLICATIONS_GROUP = "applications_group";

    private static final String APPLICATIONS_GROUP_ENABLED = "enabled";

    private static final String APPLICATIONS_GROUP_ID = "id";

    private static final String APPLICATIONS_GROUP_NAME = "name";

    private static final String APPLICATIONS_GROUP_NOTE = "note";

    private static final String APPLICATIONS_GROUP_APPLICATION_ID = "application_id";

    private static final String APPLICATIONS_GROUP_APPLICATION_IDS = "application_ids";

    private static final String APPLICATIONS_GROUP_APPLICATION_NAME = "application_name";

    private static final String APPLICATIONS_GROUP_APPLICATION_NAMES = "application_names";

    private static final String DATA_VERSION = "version";

    private static final String DATA_VERSION_1 = "1";

    private static final String ERROR_CANNOT_LOAD = "Cannot load applications group configuration, cause: {}";

    private static final String SAX_FEATURES_EXTERNAL_GENERAL_ENTITIES = "http://xml.org/sax/features/external-general-entities";

    private static final String SPLITTER_COMMAS = ",";

    private static final String SPLITTER_SPACE = " ";

    private transient List<ApplicationsGroupData> applicationsGroupDatas = null;

    private ApplicationsGroupConfig()
    {
        applicationsGroupDatas = new ArrayList<>();
    }

    public ApplicationsGroupConfig( String configFolder, String configFile )
    {
        this();
        try (FileReader fileReader = new FileReader( Paths.get( configFolder, configFile )
                                                          .toFile() ))
        {
            DocumentBuilderFactory documentBuilderFactory = DocumentBuilderFactory.newDefaultInstance();
            documentBuilderFactory.setFeature( SAX_FEATURES_EXTERNAL_GENERAL_ENTITIES, false );
            documentBuilderFactory.setValidating( false );
            DocumentBuilder documentBuilder = documentBuilderFactory.newDocumentBuilder();
            Document document = documentBuilder.parse( new InputSource( fileReader ) );
            Element rootNode = document.getDocumentElement();

            // get data version
            String dataVersion = null;
            for ( int index = 0; index < rootNode.getAttributes()
                                                 .getLength(); index++ )
            {
                Node node = rootNode.getAttributes()
                                    .item( index );
                if ( Objects.nonNull( node ) && StringUtils.equals( node.getNodeName(), DATA_VERSION ) && StringUtils.isNotBlank( node.getTextContent() ) )
                {
                    dataVersion = StringUtils.stripToEmpty( node.getTextContent() );
                }
            }

            // applications group data
            if ( StringUtils.equals( dataVersion, DATA_VERSION_1 ) )
            {
                NodeList nodeList = rootNode.getElementsByTagName( APPLICATIONS_GROUP );
                for ( int index = 0; index < nodeList.getLength(); index++ )
                {
                    Node applicationsGroupNode = nodeList.item( index );
                    if ( applicationsGroupNode.getNodeType() == Node.ELEMENT_NODE && applicationsGroupNode.hasChildNodes() )
                    {
                        ApplicationsGroupData applicationsGroupData = new ApplicationsGroupData();
                        applicationsGroupData.setId( ApplicationsGroupData.NEW_APPLICATIONS_GROUP );
                        applicationsGroupData.setApplicationDatas( new ArrayList<>() );

                        NodeList applicationsGroupNodeList = applicationsGroupNode.getChildNodes();
                        for ( int i = 0; i < applicationsGroupNodeList.getLength(); i++ )
                        {
                            Node applicationsGroupChildNode = applicationsGroupNodeList.item( i );
                            String applicationsGroupNodeText = StringUtils.stripToEmpty( applicationsGroupChildNode.getTextContent() );
                            switch ( applicationsGroupChildNode.getNodeName() )
                            {
                                case APPLICATIONS_GROUP_APPLICATION_ID:
                                    if ( StringUtils.isNumeric( applicationsGroupNodeText ) )
                                    {
                                        List<ApplicationData> datas = applicationsGroupData.getApplicationDatas();
                                        ApplicationData applicationData = new ApplicationData();
                                        applicationData.setId( Math.abs( Long.parseLong( applicationsGroupNodeText ) ) );
                                        datas.add( applicationData );
                                        applicationsGroupData.setApplicationDatas( datas );
                                    }
                                    break;
                                case APPLICATIONS_GROUP_APPLICATION_IDS:
                                    if ( StringUtils.isNumericSpace( StringUtils.replace( applicationsGroupNodeText, SPLITTER_COMMAS, SPLITTER_SPACE ) ) )
                                    {
                                        List<ApplicationData> datas = applicationsGroupData.getApplicationDatas();
                                        Arrays.asList( StringUtils.split( StringUtils.replace( applicationsGroupNodeText, SPLITTER_SPACE, SPLITTER_COMMAS ), SPLITTER_COMMAS ) )
                                              .stream()
                                              .filter( data -> StringUtils.isNotEmpty( StringUtils.stripToEmpty( data ) ) )
                                              .forEach( data -> {
                                                  ApplicationData applicationData = new ApplicationData();
                                                  applicationData.setId( Math.abs( Long.parseLong( StringUtils.strip( data ) ) ) );
                                                  datas.add( applicationData );
                                              } );
                                        applicationsGroupData.setApplicationDatas( datas );
                                    }
                                    break;
                                case APPLICATIONS_GROUP_APPLICATION_NAME:
                                    if ( StringUtils.isNotBlank( applicationsGroupNodeText ) )
                                    {
                                        List<ApplicationData> datas = applicationsGroupData.getApplicationDatas();
                                        ApplicationData applicationData = new ApplicationData();
                                        applicationData.setName( applicationsGroupNodeText );
                                        datas.add( applicationData );
                                        applicationsGroupData.setApplicationDatas( datas );
                                    }
                                    break;
                                case APPLICATIONS_GROUP_APPLICATION_NAMES:
                                    if ( StringUtils.isNotBlank( applicationsGroupNodeText ) )
                                    {
                                        List<ApplicationData> datas = applicationsGroupData.getApplicationDatas();
                                        Arrays.asList( StringUtils.split( applicationsGroupNodeText, SPLITTER_COMMAS ) )
                                              .stream()
                                              .filter( data -> StringUtils.isNotEmpty( StringUtils.stripToEmpty( data ) ) )
                                              .forEach( data -> {
                                                  ApplicationData applicationData = new ApplicationData();
                                                  applicationData.setName( StringUtils.strip( data ) );
                                                  datas.add( applicationData );
                                              } );
                                        applicationsGroupData.setApplicationDatas( datas );
                                    }
                                    break;
                                case APPLICATIONS_GROUP_ENABLED:
                                    applicationsGroupData.setEnabled( Boolean.parseBoolean( applicationsGroupNodeText ) );
                                    break;
                                case APPLICATIONS_GROUP_ID:
                                    if ( StringUtils.isNumeric( applicationsGroupNodeText ) )
                                    {
                                        applicationsGroupData.setId( Math.abs( Long.parseLong( applicationsGroupNodeText ) ) );
                                    }
                                    break;
                                case APPLICATIONS_GROUP_NAME:
                                    applicationsGroupData.setName( applicationsGroupNodeText );
                                    break;
                                case APPLICATIONS_GROUP_NOTE:
                                    applicationsGroupData.setNote( applicationsGroupNodeText );
                                    break;
                                default:
                                    // nothing to do
                                    break;
                            }
                        }

                        // add new servers group data
                        if ( StringUtils.isNotBlank( applicationsGroupData.getName() ) && CollectionUtils.isNotEmpty( applicationsGroupData.getApplicationDatas() ) )
                        {
                            applicationsGroupDatas.add( applicationsGroupData );
                        }
                    }
                }
            }
        }
        catch ( IOException | SAXException | ParserConfigurationException exception )
        {
            LOGGER.info( ERROR_CANNOT_LOAD, exception.getMessage() );
        }
    }

    public List<ApplicationsGroupData> getApplicationsGroupDatas()
    {
        return applicationsGroupDatas;
    }

}
