/**
 * Copyright (C) 2020-2022 Dušan Vejnović <vaadin@dussan.org>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.dussan.vaadin.dvirtual.config;

import java.io.FileReader;
import java.io.IOException;
import java.io.Serializable;
import java.nio.file.Paths;
import java.util.ArrayList;
import java.util.List;
import java.util.Objects;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;

import org.apache.commons.lang3.StringUtils;
import org.dussan.vaadin.dvirtual.database.data.ServerData;
import org.dussan.vaadin.dvirtual.utils.LoggerUtil;
import org.slf4j.Logger;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;

public class ServerConfig
    implements Serializable
{

    private static final long serialVersionUID = 5455925136042508191L;

    private static final Logger LOGGER = LoggerUtil.getLogger();

    private static final String DATA_VERSION = "version";

    private static final String DATA_VERSION_1 = "1";

    private static final String ERROR_CANNOT_LOAD = "Cannot load server configuration, cause: {}";

    private static final String SAX_FEATURES_EXTERNAL_GENERAL_ENTITIES = "http://xml.org/sax/features/external-general-entities";

    private static final String SERVER = "server";

    private static final String SERVER_ADDRESS = "address";

    private static final String SERVER_ENABLED = "enabled";

    private static final String SERVER_ID = "id";

    private static final String SERVER_MAX_USERS = "max_users";

    private static final String SERVER_NAME = "name";

    private static final String SERVER_NOTE = "note";

    private static final String SERVER_SSH_PORT = "ssh_port";

    private static final String SERVER_XPRA_MAX_PORT = "xpra_max_port";

    private static final String SERVER_XPRA_MIN_PORT = "xpra_min_port";

    private transient List<ServerData> serverDatas = null;

    private ServerConfig()
    {
        serverDatas = new ArrayList<>();
    }

    public ServerConfig( String configFolder, String configFile )
    {
        this();
        try (FileReader fileReader = new FileReader( Paths.get( configFolder, configFile )
                                                          .toFile() ))
        {
            DocumentBuilderFactory documentBuilderFactory = DocumentBuilderFactory.newDefaultInstance();
            documentBuilderFactory.setFeature( SAX_FEATURES_EXTERNAL_GENERAL_ENTITIES, false );
            documentBuilderFactory.setValidating( false );
            DocumentBuilder documentBuilder = documentBuilderFactory.newDocumentBuilder();
            Document document = documentBuilder.parse( new InputSource( fileReader ) );
            Element rootNode = document.getDocumentElement();

            // get data version
            String dataVersion = null;
            for ( int index = 0; index < rootNode.getAttributes()
                                                 .getLength(); index++ )
            {
                Node node = rootNode.getAttributes()
                                    .item( index );
                if ( Objects.nonNull( node ) && StringUtils.equals( node.getNodeName(), DATA_VERSION ) && StringUtils.isNotBlank( node.getTextContent() ) )
                {
                    dataVersion = StringUtils.stripToEmpty( node.getTextContent() );
                }
            }

            // server data
            if ( StringUtils.equals( dataVersion, DATA_VERSION_1 ) )
            {
                NodeList nodeList = rootNode.getElementsByTagName( SERVER );
                for ( int index = 0; index < nodeList.getLength(); index++ )
                {
                    Node serverNode = nodeList.item( index );
                    if ( serverNode.getNodeType() == Node.ELEMENT_NODE && serverNode.hasChildNodes() )
                    {
                        ServerData serverData = new ServerData();
                        serverData.setId( ServerData.NEW_SERVER );

                        NodeList serverChildNodeList = serverNode.getChildNodes();
                        for ( int i = 0; i < serverChildNodeList.getLength(); i++ )
                        {
                            Node serverChildNode = serverChildNodeList.item( i );
                            String serverNodeText = StringUtils.stripToEmpty( serverChildNode.getTextContent() );
                            switch ( serverChildNode.getNodeName() )
                            {
                                case SERVER_ADDRESS:
                                    serverData.setAddress( serverNodeText );
                                    break;
                                case SERVER_ENABLED:
                                    serverData.setEnabled( Boolean.parseBoolean( serverNodeText ) );
                                    break;
                                case SERVER_ID:
                                    if ( StringUtils.isNumeric( serverNodeText ) )
                                    {
                                        serverData.setId( Math.abs( Long.parseLong( serverNodeText ) ) );
                                    }
                                    break;
                                case SERVER_MAX_USERS:
                                    if ( StringUtils.isNumeric( serverNodeText ) )
                                    {
                                        serverData.setMaxUsers( Math.abs( Integer.parseInt( serverNodeText ) ) );
                                    }
                                    break;
                                case SERVER_NAME:
                                    serverData.setName( serverNodeText );
                                    break;
                                case SERVER_NOTE:
                                    serverData.setNote( serverNodeText );
                                    break;
                                case SERVER_SSH_PORT:
                                    if ( StringUtils.isNumeric( serverNodeText ) )
                                    {
                                        serverData.setSshPort( Math.abs( Integer.parseInt( serverNodeText ) ) );
                                    }
                                    break;
                                case SERVER_XPRA_MAX_PORT:
                                    if ( StringUtils.isNumeric( serverNodeText ) )
                                    {
                                        serverData.setXpraMaxPort( Math.abs( Integer.parseInt( serverNodeText ) ) );
                                    }
                                    break;
                                case SERVER_XPRA_MIN_PORT:
                                    if ( StringUtils.isNumeric( serverNodeText ) )
                                    {
                                        serverData.setXpraMinPort( Math.abs( Integer.parseInt( serverNodeText ) ) );
                                    }
                                    break;
                                default:
                                    // nothing to do
                                    break;
                            }
                        }

                        // add new server data
                        if ( StringUtils.isNotBlank( serverData.getAddress() ) && StringUtils.isNotBlank( serverData.getName() ) )
                        {
                            serverDatas.add( serverData );
                        }
                    }
                }
            }
        }
        catch ( IOException | SAXException | ParserConfigurationException exception )
        {
            LOGGER.info( ERROR_CANNOT_LOAD, exception.getMessage() );
        }
    }

    public List<ServerData> getServerDatas()
    {
        return serverDatas;
    }

}
