/**
 * Copyright (C) 2020-2022 Dušan Vejnović <vaadin@dussan.org>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.dussan.vaadin.dvirtual.runners;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.net.InetAddress;
import java.net.InetSocketAddress;
import java.net.Socket;
import java.net.SocketAddress;
import java.util.Objects;
import java.util.concurrent.TimeUnit;

import org.apache.commons.lang3.StringUtils;
import org.dussan.vaadin.dvirtual.client.Command;
import org.dussan.vaadin.dvirtual.i18n.Translation;
import org.dussan.vaadin.dvirtual.utils.VirtualUtil;
import org.dussan.vaadin.dvirtual.utils.EncryptionUtil;
import org.dussan.vaadin.dvirtual.utils.LoggerUtil;
import org.dussan.vaadin.dvirtual.utils.SshUtil;
import org.slf4j.Logger;

import com.jcraft.jsch.Channel;
import com.jcraft.jsch.ChannelExec;
import com.jcraft.jsch.JSchException;
import com.jcraft.jsch.Session;
import com.vaadin.flow.component.UI;
import com.vaadin.flow.component.button.Button;
import com.vaadin.flow.component.html.Label;

public class ConnectionTestRunner
    implements Runnable
{

    private static final Logger LOGGER = LoggerUtil.getLogger();

    private static final int BYTES_LENGTH = 8192;

    private static final int MAX_WAIT_INTERVAL = 5000; // milliseconds

    private static final int NO_STREAM_VALUE = -1;

    private static final int PING_INTERVAL = MAX_WAIT_INTERVAL; // milliseconds

    private static final int SLEEP_INTERVAL = 100; // milliseconds

    private static final int SOCKET_SLEEP_INTERVAL = 1000; // milliseconds

    private final Integer port;

    private final String address;

    private final Button closeButton;

    private final Label label;

    private final UI ui;

    private VirtualUtil virtualUtil = null;

    public ConnectionTestRunner( UI ui, Button closeButton, Label label, String address, Integer port, VirtualUtil virtualUtil )
    {
        this.ui = ui;
        this.closeButton = closeButton;
        this.label = label;
        this.address = address;
        this.port = port;
        this.virtualUtil = virtualUtil;
    }

    @Override
    public void run()
    {
        ui.getSession()
          .lock();

        Channel channel = null;
        Session session = null;
        try (Socket socket = new Socket())
        {
            TimeUnit.MILLISECONDS.sleep( SOCKET_SLEEP_INTERVAL );
            InetAddress host = InetAddress.getByName( address );
            boolean reachableHost = host.isReachable( PING_INTERVAL );
            if ( reachableHost )
            {
                socket.setReuseAddress( true );
                SocketAddress socketAddress = new InetSocketAddress( address, port );
                socket.connect( socketAddress, PING_INTERVAL );
                if ( socket.isConnected() )
                {
                    // start SSH session
                    session = SshUtil.createSession( virtualUtil.getUserName(), EncryptionUtil.decrypt( virtualUtil.getUserPassword(), virtualUtil.getSecretKeySpec(), virtualUtil.getSecureIV() ),
                                                     address, port );
                    session.connect();

                    // find xpra
                    byte[] tempInputStreamValue = null;
                    for ( int timer = 0; timer < MAX_WAIT_INTERVAL; timer += SLEEP_INTERVAL )
                    {
                        boolean applicationWithPath = StringUtils.contains( Command.XPRA.XPRA_APPLICATION, File.separator );
                        tempInputStreamValue = new byte[BYTES_LENGTH];
                        channel = session.openChannel( Command.SSH.EXECUTE );
                        ( (ChannelExec) channel ).setCommand( String.format( applicationWithPath ? Command.SSH.FIND_APPLICATION_WITH_PATH : Command.SSH.FIND_APPLICATION,
                                                                             Command.XPRA.XPRA_APPLICATION ) );
                        InputStream inputStream = channel.getInputStream();
                        channel.connect();
                        for ( int t = 0; t < MAX_WAIT_INTERVAL; t += SLEEP_INTERVAL )
                        {
                            TimeUnit.MILLISECONDS.sleep( SLEEP_INTERVAL );
                            if ( channel.isClosed() )
                            {
                                break;
                            }
                        }

                        int length = NO_STREAM_VALUE;
                        String inputStreamValue = null;
                        while ( ( length = inputStream.read( tempInputStreamValue, 0, BYTES_LENGTH ) ) > NO_STREAM_VALUE )
                        {
                            if ( length > 0 )
                            {
                                inputStreamValue = new String( tempInputStreamValue, 0, length );
                                if ( StringUtils.isNotBlank( inputStreamValue ) )
                                {
                                    break;
                                }
                            }
                        }
                        inputStream.close();
                        channel.disconnect();

                        if ( StringUtils.isNotBlank( inputStreamValue ) )
                        {
                            break;
                        }
                    }
                    session.disconnect();
                    ui.access( () -> label.setText( virtualUtil.getTranslation( Translation.SERVER_RESPONDING ) ) );
                }
                else
                {
                    ui.access( () -> label.setText( virtualUtil.getTranslation( Translation.SERVER_ERROR_RESPONDING ) ) );
                }
            }
            else
            {
                ui.access( () -> label.setText( virtualUtil.getTranslation( Translation.SERVER_ERROR_RESPONDING ) ) );
            }
        }
        catch ( InterruptedException | JSchException exception )
        {
            ui.access( () -> label.setText( virtualUtil.getTranslation( Translation.SERVER_ERROR_LOGIN ) ) );
            LOGGER.info( exception.getMessage() );
            Thread.currentThread()
                  .interrupt();
        }
        catch ( SecurityException exception )
        {
            ui.access( () -> label.setText( virtualUtil.getTranslation( Translation.SERVER_ERROR_SECURITY ) ) );
            LOGGER.info( exception.getMessage() );
            Thread.currentThread()
                  .interrupt();
        }
        catch ( IOException exception )
        {
            ui.access( () -> label.setText( virtualUtil.getTranslation( Translation.SERVER_ERROR_RESPONDING ) ) );
            LOGGER.info( exception.getMessage() );
            Thread.currentThread()
                  .interrupt();
        }
        finally
        {
            if ( Objects.nonNull( channel ) && channel.isConnected() )
            {
                channel.disconnect();
            }

            if ( Objects.nonNull( session ) && session.isConnected() )
            {
                session.disconnect();
            }
        }

        closeButton.setEnabled( true );
        ui.push();
        ui.getSession()
          .unlock();
    }

}
