/**
 * Copyright (C) 2020-2022 Dušan Vejnović <vaadin@dussan.org>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.dussan.vaadin.dvirtual.runners;

import java.io.IOException;
import java.io.InputStream;
import java.nio.file.Path;
import java.util.Date;
import java.util.Objects;
import java.util.concurrent.TimeUnit;

import org.apache.commons.lang3.StringUtils;
import org.dussan.vaadin.dvirtual.client.Command;
import org.dussan.vaadin.dvirtual.client.Command.XPRA;
import org.dussan.vaadin.dvirtual.database.Database;
import org.dussan.vaadin.dvirtual.database.data.ServerData;
import org.dussan.vaadin.dvirtual.database.data.UserSessionData;
import org.dussan.vaadin.dvirtual.utils.VirtualUtil;
import org.dussan.vaadin.dvirtual.utils.EncryptionUtil;
import org.dussan.vaadin.dvirtual.utils.LoggerUtil;
import org.dussan.vaadin.dvirtual.utils.SshUtil;
import org.slf4j.Logger;

import com.jcraft.jsch.Channel;
import com.jcraft.jsch.ChannelExec;
import com.jcraft.jsch.JSchException;
import com.jcraft.jsch.Session;

public class StopServerRunner
    implements Runnable
{

    private static final Logger LOGGER = LoggerUtil.getLogger();

    private static final int BYTES_LENGTH = 8192;

    private static final int MAX_WAIT_INTERVAL = 10000; // milliseconds

    private static final int NO_STREAM_VALUE = -1;

    private static final int SLEEP_INTERVAL = 100; // milliseconds

    private static final int SLEEP_XPRA_MAX_INTERVAL = 5000; // milliseconds

    private static final int SLEEP_XPRA_MIN_INTERVAL = 1000; // milliseconds

    private static final String SERVER_STOPPED = "The server is stopped.";

    private static final String SERVER_STOPPING = "Stopping the server: {}:{}";

    private static final String STRING_MINUS = "-";

    private static final String TEMP_DIR = System.getProperty( "java.io.tmpdir" );

    private final int xpraPort;

    private long userSessionId = UserSessionData.NEW_USER_SESSION;

    private Database database = null;

    private VirtualUtil virtualUtil = null;

    private ServerData serverData = null;

    public StopServerRunner( ServerData serverData, int xpraPort, VirtualUtil virtualUtil )
    {
        this.serverData = serverData;
        this.xpraPort = xpraPort;
        this.virtualUtil = virtualUtil;
    }

    public StopServerRunner( Database database, long userSessionId, ServerData serverData, int xpraPort, VirtualUtil virtualUtil )
    {
        this.database = database;
        this.userSessionId = userSessionId;
        this.serverData = serverData;
        this.xpraPort = xpraPort;
        this.virtualUtil = virtualUtil;
    }

    @Override
    public void run()
    {
        // running application will be closed
        virtualUtil.setRunningApplicationInClosingProcess( true );

        Channel channel = null;
        Session session = null;
        try
        {
            LOGGER.info( SERVER_STOPPING, serverData.getAddress(), xpraPort );

            // start SSH session
            session = SshUtil.createSession( virtualUtil.getUserName(), EncryptionUtil.decrypt( virtualUtil.getUserPassword(), virtualUtil.getSecretKeySpec(), virtualUtil.getSecureIV() ),
                                             serverData.getAddress(), serverData.getSshPort() );
            session.connect();

            // get status of running XPRA server
            Date startDate = new Date();
            boolean serverRunning = true;
            byte[] temporaryValue = new byte[BYTES_LENGTH];
            channel = session.openChannel( Command.SSH.EXECUTE );
            ( (ChannelExec) channel ).setCommand( Command.XPRA.LIST );
            InputStream inputStream = channel.getInputStream();
            channel.connect();
            for ( int t = 0; t < MAX_WAIT_INTERVAL; t += SLEEP_INTERVAL )
            {
                TimeUnit.MILLISECONDS.sleep( SLEEP_INTERVAL );
                if ( channel.isClosed() )
                {
                    break;
                }
            }
            int length = NO_STREAM_VALUE;
            while ( ( length = inputStream.read( temporaryValue, 0, BYTES_LENGTH ) ) > NO_STREAM_VALUE )
            {
                if ( length > 0 )
                {
                    String status = new String( temporaryValue, 0, length );
                    serverRunning = StringUtils.contains( status, String.format( Command.XPRA.LIST_STATUS, xpraPort ) );
                    break;
                }
            }
            inputStream.close();
            channel.disconnect();

            // stop XPRA server
            if ( serverRunning )
            {
                channel = session.openChannel( Command.SSH.EXECUTE );
                ( (ChannelExec) channel ).setCommand( String.format( Command.XPRA.STOP, xpraPort ) );
                channel.connect();
                for ( int timer = 0; timer < MAX_WAIT_INTERVAL; timer += SLEEP_INTERVAL )
                {
                    TimeUnit.MILLISECONDS.sleep( SLEEP_INTERVAL );
                    if ( channel.isClosed() )
                    {
                        break;
                    }
                }
                channel.disconnect();

                // get status of stopped XPRA server
                boolean serverStopped = false;
                temporaryValue = new byte[BYTES_LENGTH];
                for ( int timer = 0; timer < MAX_WAIT_INTERVAL; timer += SLEEP_INTERVAL )
                {
                    channel = session.openChannel( Command.SSH.EXECUTE );
                    ( (ChannelExec) channel ).setCommand( Command.XPRA.LIST );
                    inputStream = channel.getInputStream();
                    channel.connect();
                    for ( int t = 0; t < MAX_WAIT_INTERVAL; t += SLEEP_INTERVAL )
                    {
                        TimeUnit.MILLISECONDS.sleep( SLEEP_INTERVAL );
                        if ( channel.isClosed() )
                        {
                            break;
                        }
                    }

                    length = NO_STREAM_VALUE;
                    while ( ( length = inputStream.read( temporaryValue, 0, BYTES_LENGTH ) ) > NO_STREAM_VALUE )
                    {
                        if ( length > 0 )
                        {
                            String status = new String( temporaryValue, 0, length );
                            if ( !StringUtils.contains( status, String.format( Command.XPRA.LIST_STATUS, xpraPort ) ) )
                            {
                                serverStopped = true;
                                break;
                            }
                        }
                    }
                    inputStream.close();
                    channel.disconnect();

                    if ( serverStopped )
                    {
                        break;
                    }
                }
            }
            Date stopDate = new Date();
            long stopStartDateDiff = ( stopDate.getTime() - startDate.getTime() );
            stopStartDateDiff = Math.max( stopStartDateDiff, SLEEP_XPRA_MIN_INTERVAL );
            stopStartDateDiff = Math.min( stopStartDateDiff, SLEEP_XPRA_MAX_INTERVAL );
            TimeUnit.MILLISECONDS.sleep( stopStartDateDiff );

            // Xpra session
            temporaryValue = new byte[BYTES_LENGTH];
            channel = session.openChannel( Command.SSH.EXECUTE );
            ( (ChannelExec) channel ).setCommand( Command.XPRA.LIST );
            inputStream = channel.getInputStream();
            channel.connect();
            for ( int t = 0; t < MAX_WAIT_INTERVAL; t += SLEEP_INTERVAL )
            {
                TimeUnit.MILLISECONDS.sleep( SLEEP_INTERVAL );
                if ( channel.isClosed() )
                {
                    break;
                }
            }
            length = NO_STREAM_VALUE;
            String status = StringUtils.EMPTY;
            while ( ( length = inputStream.read( temporaryValue, 0, BYTES_LENGTH ) ) > NO_STREAM_VALUE )
            {
                if ( length > 0 )
                {
                    status = new String( temporaryValue, 0, length );
                }
            }
            inputStream.close();
            channel.disconnect();
            if ( StringUtils.containsIgnoreCase( status, Command.XPRA.NO_XPRA_SESSION ) )
            {
                // stop Xpra pulseaudio
                channel = session.openChannel( Command.SSH.EXECUTE );
                ( (ChannelExec) channel ).setCommand( String.format( Command.SSH.KILL_XPRA_PULSEAUDIO, virtualUtil.getUserName() ) );
                channel.connect();
                for ( int t = 0; t < MAX_WAIT_INTERVAL; t += SLEEP_INTERVAL )
                {
                    TimeUnit.MILLISECONDS.sleep( SLEEP_INTERVAL );
                    if ( channel.isClosed() )
                    {
                        break;
                    }
                }
                channel.disconnect();

                // delete xpra fakexinerama files
                channel = session.openChannel( Command.SSH.EXECUTE );
                ( (ChannelExec) channel ).setCommand( Command.SSH.DELETE_FAKEXINERAMA_FILES );
                channel.connect();
                for ( int timer = 0; timer < MAX_WAIT_INTERVAL; timer += SLEEP_INTERVAL )
                {
                    TimeUnit.MILLISECONDS.sleep( SLEEP_INTERVAL );
                    if ( channel.isClosed() )
                    {
                        break;
                    }
                }
                channel.disconnect();
            }

            // delete xpra folder
            String tmpDir = StringUtils.join( Path.of( TEMP_DIR, XPRA.XPRA_APPLICATION ), STRING_MINUS, virtualUtil.getUserName(), STRING_MINUS, xpraPort );
            channel = session.openChannel( Command.SSH.EXECUTE );
            ( (ChannelExec) channel ).setCommand( String.format( Command.SSH.DELETE_XPRA_FOLDER, tmpDir ) );
            channel.connect();
            for ( int timer = 0; timer < MAX_WAIT_INTERVAL; timer += SLEEP_INTERVAL )
            {
                TimeUnit.MILLISECONDS.sleep( SLEEP_INTERVAL );
                if ( channel.isClosed() )
                {
                    break;
                }
            }
            channel.disconnect();
            session.disconnect();

            // remove user session data from database
            if ( Objects.nonNull( database ) )
            {
                UserSessionData userSessionData =
                    userSessionId != UserSessionData.NEW_USER_SESSION ? database.getUserSession( userSessionId ) : database.getUserSession( serverData.getId(), xpraPort );
                if ( Objects.nonNull( userSessionData ) )
                {
                    database.removeUserSession( userSessionData );
                    database.removeUserSessionApplication( userSessionData );
                }
            }
        }
        catch ( IllegalThreadStateException | InterruptedException | IOException | JSchException exception )
        {
            Thread.currentThread()
                  .interrupt();
        }
        finally
        {
            if ( Objects.nonNull( channel ) && channel.isConnected() )
            {
                channel.disconnect();
            }

            if ( Objects.nonNull( session ) && session.isConnected() )
            {
                session.disconnect();
            }
        }

        // running application is closed
        LOGGER.info( SERVER_STOPPED );
        virtualUtil.setRunningApplicationInClosingProcess( false );
    }

}
