/**
 * Copyright (C) 2020-2022 Dušan Vejnović <vaadin@dussan.org>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.dussan.vaadin.dvirtual.ui.file;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.Objects;
import java.util.concurrent.TimeUnit;

import org.apache.commons.collections4.CollectionUtils;
import org.apache.commons.lang3.StringUtils;
import org.dussan.vaadin.dvirtual.client.Command;
import org.dussan.vaadin.dvirtual.database.columns.FileColumns;
import org.dussan.vaadin.dvirtual.database.data.FileData;
import org.dussan.vaadin.dvirtual.database.data.UserData;
import org.dussan.vaadin.dvirtual.database.data.UserSessionData;
import org.dussan.vaadin.dvirtual.i18n.Translation;
import org.dussan.vaadin.dvirtual.utils.VirtualUtil;
import org.dussan.vaadin.dvirtual.utils.EncryptionUtil;
import org.dussan.vaadin.dvirtual.utils.SshUtil;
import org.dussan.vaadin.dvirtual.utils.dialog.DialogEditUtil;
import org.dussan.vaadin.dvirtual.utils.dialog.DialogInfoUtil;

import com.jcraft.jsch.Channel;
import com.jcraft.jsch.ChannelExec;
import com.jcraft.jsch.JSchException;
import com.jcraft.jsch.Session;
import com.vaadin.flow.component.button.Button;
import com.vaadin.flow.component.formlayout.FormLayout;
import com.vaadin.flow.component.formlayout.FormLayout.ResponsiveStep;
import com.vaadin.flow.component.formlayout.FormLayout.ResponsiveStep.LabelsPosition;
import com.vaadin.flow.component.grid.ColumnTextAlign;
import com.vaadin.flow.component.grid.Grid;
import com.vaadin.flow.component.grid.Grid.SelectionMode;
import com.vaadin.flow.component.html.Div;
import com.vaadin.flow.component.html.Span;
import com.vaadin.flow.component.orderedlayout.FlexComponent.JustifyContentMode;
import com.vaadin.flow.component.orderedlayout.HorizontalLayout;
import com.vaadin.flow.component.orderedlayout.VerticalLayout;
import com.vaadin.flow.component.textfield.TextField;

public abstract class DownloadFileDialog
    extends DialogEditUtil
{

    private static final long serialVersionUID = -3122125982262411686L;

    private static final int BYTES_LENGTH = 8192;

    private static final int COLUMN_DAY = 3;

    private static final int COLUMN_MONTH = 2;

    private static final int COLUMN_NAME = 5;

    private static final int COLUMN_SIZE = 1;

    private static final int COLUMN_YEAR_TIME = 4;

    private static final int FIRST_ELEMENT = 0;

    private static final int FORM_COLUMNS = 1;

    private static final int MAX_WAIT_INTERVAL = 30000; // milliseconds

    private static final int SLEEP_INTERVAL = 50; // milliseconds

    private static final String ATTRIBUTE_CLASS_BUTTON_VALUE = "primary";

    private static final String ATTRIBUTE_THEME = "theme";

    private static final String ATTRIBUTE_THEME_VALUE = "dark";

    private static final String CLASS_DIALOG = "edit-dialog";

    private static final String CLASS_GRID_CELL_THEME = "grid-cell-theme";

    private static final String FILE_PARENT = StringUtils.join( File.separator, ".." );

    private static final String FILE_SEPARATOR = File.separator;

    private static final String FORM_MINIMAL_WIDTH = "0";

    private static final String INDICATOR_FOLDER = "d";

    private static final String LINE_SEPARATOR = System.getProperty( "line.separator" );

    private static final String STRING_COMMA_VALUE = ",";

    private static final String STRING_SPACE_VALUE = " ";

    private static final String STYLE_DISPLAY = "display";

    private static final String STYLE_DISPLAY_GRID_VALUE = "grid";

    private static final String STYLE_MARGIN_BOTTOM = "margin-bottom";

    private static final String STYLE_MARGIN_LEFT = "margin-left";

    private static final String STYLE_MARGIN_RIGHT = "margin-right";

    private static final String STYLE_MARGIN_VALUE = "0px";

    private static final String STYLE_PADDING_BOTTOM = "padding-bottom";

    private static final String STYLE_PADDING_LEFT = "padding-left";

    private static final String STYLE_PADDING_RIGHT = "padding-right";

    private static final String STYLE_PADDING_TOP = "padding-top";

    private static final String STYLE_PADDING_VALUE = "0px";

    private static final String STYLE_TEXT_ALIGN = "text-align";

    private static final String STYLE_TEXT_ALIGN_VALUE = "center";

    private static final String STYLE_WIDTH = "width";

    private static final String STYLE_WIDTH_100PERCENT_VALUE = "100%";

    private static final String STYLE_WIDTH_100PX_VALUE = "100px";

    private static final String STYLE_WIDTH_150PX_VALUE = "150px";

    private static final String STYLE_WIDTH_70PX_VALUE = "70px";

    private static final String STYLE_WIDTH_80PX_VALUE = "80px";

    private static final String STYLE_WIDTH_900PX_VALUE = "900px";

    private transient List<FileData> files = null;

    private UserSessionData userSessionData = null;

    private Grid<FileData> fileGrid = null;

    private TextField currentFolderTextField = null;

    private VirtualUtil virtualUtil = null;

    public DownloadFileDialog( UserSessionData userSessionData, VirtualUtil virtualUtil )
    {
        this.userSessionData = userSessionData;
        this.virtualUtil = virtualUtil;
        setWidth( STYLE_WIDTH_900PX_VALUE );

        // close buttons
        Button closeButton = new Button( virtualUtil.getTranslation( Translation.BUTTON_CLOSE ), event -> close() );
        closeButton.setDisableOnClick( true );
        closeButton.getElement()
                   .setAttribute( ATTRIBUTE_THEME, ATTRIBUTE_THEME_VALUE );

        // download the file buttons
        Button dowloadFileButton = new Button( virtualUtil.getTranslation( Translation.BUTTON_DOWNLOAD ), event -> {
            List<FileData> fileDatas = new ArrayList<>( fileGrid.getSelectedItems() );
            if ( CollectionUtils.isNotEmpty( fileDatas ) )
            {
                fileDownload( fileDatas.get( FIRST_ELEMENT ) );
            }
        } );
        dowloadFileButton.setDisableOnClick( true );
        dowloadFileButton.setEnabled( false );
        dowloadFileButton.addClassName( ATTRIBUTE_CLASS_BUTTON_VALUE );
        dowloadFileButton.getElement()
                         .setAttribute( ATTRIBUTE_THEME, ATTRIBUTE_THEME_VALUE );
        dowloadFileButton.getStyle()
                         .set( STYLE_MARGIN_LEFT, STYLE_MARGIN_VALUE );

        // enter the folder buttons
        Button enterFolderButton = new Button( virtualUtil.getTranslation( Translation.BUTTON_FOLDER_ENTER ), event -> {
            List<FileData> fileDatas = new ArrayList<>( fileGrid.getSelectedItems() );
            if ( CollectionUtils.isNotEmpty( fileDatas ) )
            {
                enterFolder( fileDatas.get( FIRST_ELEMENT ) );
            }
        } );
        enterFolderButton.setDisableOnClick( true );
        enterFolderButton.setEnabled( false );
        enterFolderButton.getElement()
                         .setAttribute( ATTRIBUTE_THEME, ATTRIBUTE_THEME_VALUE );

        // buttons layout
        HorizontalLayout buttonsLayout = new HorizontalLayout( dowloadFileButton, enterFolderButton, closeButton );
        buttonsLayout.setJustifyContentMode( JustifyContentMode.CENTER );
        buttonsLayout.setMargin( true );
        buttonsLayout.setPadding( true );
        buttonsLayout.setSpacing( true );
        buttonsLayout.setWidth( STYLE_WIDTH_100PERCENT_VALUE );
        buttonsLayout.getStyle()
                     .set( STYLE_MARGIN_BOTTOM, STYLE_MARGIN_VALUE )
                     .set( STYLE_MARGIN_LEFT, STYLE_MARGIN_VALUE )
                     .set( STYLE_MARGIN_RIGHT, STYLE_MARGIN_VALUE )
                     .set( STYLE_PADDING_BOTTOM, STYLE_PADDING_VALUE )
                     .set( STYLE_PADDING_LEFT, STYLE_PADDING_VALUE )
                     .set( STYLE_PADDING_RIGHT, STYLE_PADDING_VALUE );

        currentFolderTextField = new TextField( virtualUtil.getTranslation( Translation.FIELD_FOLDER ) );
        currentFolderTextField.setSizeFull();
        currentFolderTextField.setValue( getHomeFolder() );
        currentFolderTextField.setReadOnly( true );
        currentFolderTextField.getStyle()
                              .set( STYLE_PADDING_TOP, STYLE_PADDING_VALUE );

        files = new ArrayList<>();
        fileGrid = new Grid<>();
        fileGrid.setItems( files );
        fileGrid.setId( CLASS_GRID_CELL_THEME );
        fileGrid.setSelectionMode( SelectionMode.SINGLE );
        fileGrid.getStyle()
                .set( STYLE_DISPLAY, STYLE_DISPLAY_GRID_VALUE )
                .set( STYLE_WIDTH, STYLE_WIDTH_100PERCENT_VALUE );
        fileGrid.addItemDoubleClickListener( event -> {
            if ( event.getItem()
                      .getName()
                      .startsWith( FILE_SEPARATOR ) )
            {
                enterFolder( event.getItem() );
            }
            else
            {
                fileDownload( event.getItem() );
            }
        } );
        fileGrid.addSelectionListener( event ->

        {
            dowloadFileButton.setEnabled( false );
            enterFolderButton.setEnabled( false );
            List<FileData> fileDatas = new ArrayList<>( event.getAllSelectedItems() );
            if ( CollectionUtils.isNotEmpty( fileDatas ) )
            {
                FileData fileData = fileDatas.get( FIRST_ELEMENT );
                dowloadFileButton.setEnabled( !StringUtils.startsWith( fileData.getName(), FILE_SEPARATOR ) );
                enterFolderButton.setEnabled( StringUtils.startsWith( fileData.getName(), FILE_SEPARATOR ) );
            }
        } );

        // file: name
        fileGrid.addColumn( FileData::getName )
                .setFlexGrow( 1 )
                .setKey( FileColumns.NAME.getName() )
                .setResizable( true )
                .setSortable( false )
                .setHeader( virtualUtil.getTranslation( FileColumns.NAME.getValueForTranslation() ) );

        // file: size
        fileGrid.addColumn( FileData::getSize )
                .setFlexGrow( 0 )
                .setKey( FileColumns.SIZE.getName() )
                .setResizable( true )
                .setSortable( false )
                .setTextAlign( ColumnTextAlign.END )
                .setWidth( STYLE_WIDTH_150PX_VALUE )
                .setHeader( virtualUtil.getTranslation( FileColumns.SIZE.getValueForTranslation() ) );

        // file: month
        Grid.Column<FileData> monthColumn = fileGrid.addColumn( FileData::getMonth )
                                                    .setFlexGrow( 0 )
                                                    .setKey( FileColumns.MONTH.getName() )
                                                    .setResizable( false )
                                                    .setSortable( false )
                                                    .setWidth( STYLE_WIDTH_80PX_VALUE )
                                                    .setHeader( virtualUtil.getTranslation( FileColumns.MONTH.getValueForTranslation() ) );

        // file: day
        Grid.Column<FileData> dayColumn = fileGrid.addColumn( FileData::getDay )
                                                  .setFlexGrow( 0 )
                                                  .setKey( FileColumns.DAY.getName() )
                                                  .setResizable( false )
                                                  .setSortable( false )
                                                  .setTextAlign( ColumnTextAlign.END )
                                                  .setWidth( STYLE_WIDTH_70PX_VALUE )
                                                  .setHeader( virtualUtil.getTranslation( FileColumns.DAY.getValueForTranslation() ) );

        // file: time
        Grid.Column<FileData> timeColumn = fileGrid.addColumn( FileData::getYearOrTime )
                                                   .setFlexGrow( 0 )
                                                   .setKey( FileColumns.TIME.getName() )
                                                   .setResizable( false )
                                                   .setSortable( false )
                                                   .setTextAlign( ColumnTextAlign.END )
                                                   .setWidth( STYLE_WIDTH_100PX_VALUE )
                                                   .setHeader( virtualUtil.getTranslation( FileColumns.TIME.getValueForTranslation() ) );

        Div modificationTimeDiv = new Div( new Span( virtualUtil.getTranslation( Translation.FIELD_MODIFICATION ) ) );
        modificationTimeDiv.setSizeFull();
        modificationTimeDiv.getStyle()
                           .set( STYLE_TEXT_ALIGN, STYLE_TEXT_ALIGN_VALUE );
        fileGrid.prependHeaderRow()
                .join( monthColumn, dayColumn, timeColumn )
                .setComponent( modificationTimeDiv );

        List<String> fileList = new ArrayList<>( Arrays.asList( getFoldersWithFiles( currentFolderTextField.getValue() ) ) );
        fillFoldersWithFilesGrid( fileList );

        VerticalLayout serverGridLayout = new VerticalLayout( fileGrid );
        serverGridLayout.setMargin( false );
        serverGridLayout.setPadding( false );
        serverGridLayout.setSpacing( false );
        serverGridLayout.setSizeUndefined();
        serverGridLayout.setWidth( STYLE_WIDTH_100PERCENT_VALUE );

        // form layout
        setFormLayout( new FormLayout( currentFolderTextField, fileGrid, buttonsLayout ) );
        getFormLayout().setClassName( CLASS_DIALOG );
        getFormLayout().setResponsiveSteps( new ResponsiveStep( FORM_MINIMAL_WIDTH, FORM_COLUMNS, LabelsPosition.TOP ) );
        getFormLayout().setWidth( STYLE_WIDTH_100PERCENT_VALUE );
        add( getFormLayout() );
    }

    public abstract void downloadingFile( String downloadedFileName );

    private void enterFolder( FileData fileData )
    {
        if ( fileData.getName()
                     .startsWith( FILE_SEPARATOR ) )
        {
            String newCurrentFolder = StringUtils.join( currentFolderTextField.getValue(), fileData.getName() );
            currentFolderTextField.setReadOnly( false );
            if ( newCurrentFolder.endsWith( FILE_PARENT ) )
            {
                List<String> datas = new ArrayList<>( Arrays.asList( newCurrentFolder.split( FILE_SEPARATOR ) ) );
                datas = datas.subList( 0, Math.max( 1, datas.size() - 2 ) );
                currentFolderTextField.setValue( StringUtils.join( FILE_SEPARATOR, StringUtils.join( datas, FILE_SEPARATOR ) ) );
            }
            else
            {
                currentFolderTextField.setValue( newCurrentFolder );
            }
            currentFolderTextField.setValue( StringUtils.replace( currentFolderTextField.getValue(), StringUtils.join( FILE_SEPARATOR, FILE_SEPARATOR ), FILE_SEPARATOR ) );
            currentFolderTextField.setReadOnly( true );

            List<String> foldersWithFilesList = new ArrayList<>( Arrays.asList( getFoldersWithFiles( newCurrentFolder ) ) );
            fillFoldersWithFilesGrid( foldersWithFilesList );
        }
    }

    private void fileDownload( FileData fileData )
    {
        UserData userData = userSessionData.getUserData();
        if ( Long.parseLong( fileData.getSize() ) <= userData.getMaxFileSizeInBytes() )
        {
            String downloadedFileName = StringUtils.join( currentFolderTextField.getValue(), FILE_SEPARATOR, fileData.getName() );
            downloadingFile( downloadedFileName );
            fileGrid.deselectAll();
        }
        else
        {
            new DialogInfoUtil( String.format( virtualUtil.getTranslation( Translation.DATA_ERROR_SIZE_FILE ), userData.getMaxFileSize() ),
                                virtualUtil.getTranslation( Translation.BUTTON_CLOSE ) ).open();
        }
    }

    private void fillFoldersWithFilesGrid( List<String> foldersWithFiles )
    {
        // refill grid with the new data
        fileGrid.deselectAll();
        files.clear();
        for ( String folderWithFiles : foldersWithFiles )
        {
            List<String> datas = new ArrayList<>( Arrays.asList( StringUtils.split( StringUtils.stripToEmpty( folderWithFiles ) ) ) );
            datas.removeIf( String::isEmpty );
            if ( datas.size() > 2 )
            {
                if ( datas.get( COLUMN_SIZE )
                          .endsWith( STRING_COMMA_VALUE ) )
                {
                    datas.set( COLUMN_SIZE, StringUtils.join( datas.get( COLUMN_SIZE ), datas.get( COLUMN_MONTH ) ) );
                    datas.remove( COLUMN_MONTH );
                }

                boolean folderIndicator = StringUtils.startsWith( datas.get( FIRST_ELEMENT ), INDICATOR_FOLDER );
                FileData fileData = new FileData();
                if ( folderIndicator )
                {
                    fileData.setName( StringUtils.join( folderIndicator ? FILE_SEPARATOR : STRING_SPACE_VALUE, StringUtils.stripToEmpty( datas.get( COLUMN_NAME ) ) ) );
                    fileData.setSize( StringUtils.stripToEmpty( datas.get( COLUMN_SIZE ) ) );
                    fileData.setMonth( StringUtils.stripToEmpty( datas.get( COLUMN_MONTH ) ) );
                    fileData.setDay( StringUtils.stripToEmpty( datas.get( COLUMN_DAY ) ) );
                    fileData.setYearOrTime( StringUtils.stripToEmpty( datas.get( COLUMN_YEAR_TIME ) ) );
                }
                else
                {
                    StringBuilder stringBuilder = new StringBuilder();
                    for ( int index = COLUMN_NAME; index < datas.size(); index++ )
                    {
                        stringBuilder.append( StringUtils.join( datas.get( index ), STRING_SPACE_VALUE ) );
                    }
                    fileData.setName( StringUtils.stripToEmpty( stringBuilder.toString() ) );
                    fileData.setSize( StringUtils.stripToEmpty( datas.get( COLUMN_SIZE ) ) );
                    fileData.setMonth( StringUtils.stripToEmpty( datas.get( COLUMN_MONTH ) ) );
                    fileData.setDay( StringUtils.stripToEmpty( datas.get( COLUMN_DAY ) ) );
                    fileData.setYearOrTime( StringUtils.stripToEmpty( datas.get( COLUMN_YEAR_TIME ) ) );
                }
                files.add( fileData );
            }
        }

        // current folder name is not equal root folder, so add parent folder (/..)
        if ( !StringUtils.equals( currentFolderTextField.getValue(), FILE_SEPARATOR ) )
        {
            files.add( 0, new FileData( FILE_PARENT, StringUtils.EMPTY, StringUtils.EMPTY, StringUtils.EMPTY, StringUtils.EMPTY ) );
        }

        // refresh grid
        fileGrid.getDataProvider()
                .refreshAll();
        fileGrid.scrollToStart();
    }

    private String getHomeFolder()
    {
        String inputStreamValue = null;
        Channel channel = null;
        Session session = null;
        InputStream inputStream = null;
        byte[] tempInputStreamValue = null;
        StringBuilder stringBuilder = new StringBuilder();

        if ( Objects.nonNull( userSessionData ) && Objects.nonNull( virtualUtil ) )
        {
            try
            {
                // start SSH session
                session = SshUtil.createSession( virtualUtil.getUserName(), EncryptionUtil.decrypt( virtualUtil.getUserPassword(), virtualUtil.getSecretKeySpec(), virtualUtil.getSecureIV() ),
                                                 userSessionData.getServerData()
                                                                .getAddress(),
                                                 userSessionData.getServerData()
                                                                .getSshPort() );
                session.connect();

                // get user home
                tempInputStreamValue = new byte[BYTES_LENGTH];
                channel = session.openChannel( Command.SSH.EXECUTE );
                ( (ChannelExec) channel ).setCommand( String.format( Command.SSH.USER_HOME, userSessionData.getUserData()
                                                                                                           .getUserName() ) );
                inputStream = channel.getInputStream();
                channel.connect();
                for ( int timer = 0; timer < MAX_WAIT_INTERVAL; timer += SLEEP_INTERVAL )
                {
                    TimeUnit.MILLISECONDS.sleep( SLEEP_INTERVAL );
                    if ( channel.isClosed() )
                    {
                        break;
                    }
                }
                while ( inputStream.available() > 0 )
                {
                    int length = inputStream.read( tempInputStreamValue, 0, BYTES_LENGTH );
                    if ( length > 0 )
                    {
                        inputStreamValue = new String( tempInputStreamValue, 0, length );
                        if ( StringUtils.isNotBlank( inputStreamValue ) )
                        {
                            stringBuilder.append( StringUtils.stripToEmpty( inputStreamValue ) );
                            break;
                        }
                    }
                }
                inputStream.close();
                channel.disconnect();
                session.disconnect();
            }
            catch ( IllegalThreadStateException | InterruptedException | IOException | JSchException exception )
            {
                Thread.currentThread()
                      .interrupt();
            }
            finally
            {
                if ( Objects.nonNull( channel ) && channel.isConnected() )
                {
                    channel.disconnect();
                }

                if ( Objects.nonNull( session ) && session.isConnected() )
                {
                    session.disconnect();
                }
            }
        }

        return stringBuilder.toString();
    }

    private String[] getFoldersWithFiles( String folder )
    {
        String inputStreamValue = null;
        Channel channel = null;
        Session session = null;
        InputStream inputStream = null;
        byte[] tempInputStreamValue = null;
        StringBuilder stringBuilder = new StringBuilder();

        if ( Objects.nonNull( userSessionData ) && Objects.nonNull( virtualUtil ) )
        {
            try
            {
                // start SSH session
                session = SshUtil.createSession( virtualUtil.getUserName(), EncryptionUtil.decrypt( virtualUtil.getUserPassword(), virtualUtil.getSecretKeySpec(), virtualUtil.getSecureIV() ),
                                                 userSessionData.getServerData()
                                                                .getAddress(),
                                                 userSessionData.getServerData()
                                                                .getSshPort() );
                session.connect();

                // get user home
                tempInputStreamValue = new byte[BYTES_LENGTH];
                channel = session.openChannel( Command.SSH.EXECUTE );
                ( (ChannelExec) channel ).setCommand( String.format( Command.SSH.LIST_FOLDER, folder ) );
                inputStream = channel.getInputStream();
                channel.connect();
                for ( int timer = 0; timer < MAX_WAIT_INTERVAL; timer += SLEEP_INTERVAL )
                {
                    TimeUnit.MILLISECONDS.sleep( SLEEP_INTERVAL );
                    if ( channel.isClosed() )
                    {
                        break;
                    }
                }
                while ( inputStream.available() > 0 )
                {
                    int length = inputStream.read( tempInputStreamValue, 0, BYTES_LENGTH );
                    if ( length > 0 )
                    {
                        inputStreamValue = new String( tempInputStreamValue, 0, length );
                        if ( !StringUtils.isBlank( inputStreamValue ) )
                        {
                            stringBuilder.append( inputStreamValue );
                        }
                    }
                }
                inputStream.close();
                channel.disconnect();
                session.disconnect();
            }
            catch ( IllegalThreadStateException | InterruptedException | IOException | JSchException exception )
            {
                Thread.currentThread()
                      .interrupt();
            }
            finally
            {
                if ( Objects.nonNull( channel ) && channel.isConnected() )
                {
                    channel.disconnect();
                }

                if ( Objects.nonNull( session ) && session.isConnected() )
                {
                    session.disconnect();
                }
            }
        }

        return StringUtils.split( stringBuilder.toString(), LINE_SEPARATOR );
    }

}
