/**
 * Copyright (C) 2020-2022 Dušan Vejnović <vaadin@dussan.org>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.dussan.vaadin.dvirtual.ui.servergroups;

import java.util.ArrayList;
import java.util.EnumMap;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.stream.Collectors;
import java.util.stream.LongStream;

import org.apache.commons.collections4.CollectionUtils;
import org.apache.commons.lang3.StringUtils;
import org.dussan.vaadin.dvirtual.database.Database;
import org.dussan.vaadin.dvirtual.database.columns.ServerColumns;
import org.dussan.vaadin.dvirtual.database.data.ServerData;
import org.dussan.vaadin.dvirtual.database.data.ServersGroupData;
import org.dussan.vaadin.dvirtual.i18n.Translation;
import org.dussan.vaadin.dvirtual.utils.VirtualUtil;
import org.dussan.vaadin.dvirtual.utils.dialog.DialogEditUtil;
import org.dussan.vaadin.dvirtual.utils.dialog.DialogInfoUtil;

import com.vaadin.flow.component.button.Button;
import com.vaadin.flow.component.checkbox.Checkbox;
import com.vaadin.flow.component.combobox.ComboBox;
import com.vaadin.flow.component.formlayout.FormLayout;
import com.vaadin.flow.component.formlayout.FormLayout.ResponsiveStep;
import com.vaadin.flow.component.formlayout.FormLayout.ResponsiveStep.LabelsPosition;
import com.vaadin.flow.component.grid.Grid;
import com.vaadin.flow.component.grid.Grid.SelectionMode;
import com.vaadin.flow.component.grid.HeaderRow;
import com.vaadin.flow.component.html.Div;
import com.vaadin.flow.component.html.Label;
import com.vaadin.flow.component.icon.Icon;
import com.vaadin.flow.component.icon.VaadinIcon;
import com.vaadin.flow.component.orderedlayout.FlexComponent.Alignment;
import com.vaadin.flow.component.orderedlayout.FlexComponent.JustifyContentMode;
import com.vaadin.flow.component.orderedlayout.HorizontalLayout;
import com.vaadin.flow.component.orderedlayout.VerticalLayout;
import com.vaadin.flow.component.textfield.TextField;
import com.vaadin.flow.data.binder.Binder;
import com.vaadin.flow.data.provider.DataProvider;
import com.vaadin.flow.data.provider.SortDirection;
import com.vaadin.flow.data.provider.SortOrder;
import com.vaadin.flow.data.renderer.ComponentRenderer;
import com.vaadin.flow.data.validator.StringLengthValidator;
import com.vaadin.flow.data.value.ValueChangeMode;

public abstract class ServersGroupEditDialog
    extends DialogEditUtil
{

    private static final long serialVersionUID = -8470555752877541909L;

    private static final int FORM_COLUMNS = 1;

    private static final int NO_LIMITS = Integer.MAX_VALUE;

    private static final int VALUE_CHANGE_TIMEOUT = 500;

    private static final int VISIBLE_ROWS = 50;

    private static final String ATTRIBUTE_CLASS_BUTTON_VALUE = "primary";

    private static final String ATTRIBUTE_THEME = "theme";

    private static final String ATTRIBUTE_THEME_VALUE = "dark";

    private static final String CLASS_DIALOG = "edit-dialog";

    private static final String CLASS_GRID_CELL_ACTIVE = "active";

    private static final String CLASS_GRID_CELL_NOT_ACTIVE = StringUtils.EMPTY;

    private static final String CLASS_GRID_CELL_THEME = "grid-cell-theme";

    private static final String CLASS_GRID_LABEL = "grid-label";

    private static final String FORM_MINIMAL_WIDTH = "0";

    private static final String STYLE_DISPLAY = "display";

    private static final String STYLE_DISPLAY_GRID_VALUE = "grid";

    private static final String STYLE_GRID_COMBOBOX = "grid-combobox";

    private static final String STYLE_HEIGHT = "height";

    private static final String STYLE_HEIGHT_16PX_VALUE = "16px";

    private static final String STYLE_MARGIN_BOTTOM = "margin-bottom";

    private static final String STYLE_MARGIN_LEFT = "margin-left";

    private static final String STYLE_MARGIN_RIGHT = "margin-right";

    private static final String STYLE_MARGIN_VALUE = "0px";

    private static final String STYLE_PADDING_14PX_VALUE = "14px";

    private static final String STYLE_PADDING_BOTTOM = "padding-bottom";

    private static final String STYLE_PADDING_LEFT = "padding-left";

    private static final String STYLE_PADDING_RIGHT = "padding-right";

    private static final String STYLE_PADDING_TOP = "padding-top";

    private static final String STYLE_PADDING_VALUE = "0px";

    private static final String STYLE_TEXT_ALIGN = "text-align";

    private static final String STYLE_TEXT_ALIGN_VALUE = "center";

    private static final String STYLE_WIDTH = "width";

    private static final String STYLE_WIDTH_100PERCENT_VALUE = "100%";

    private static final String STYLE_WIDTH_130PX_VALUE = "130px";

    private static final String STYLE_WIDTH_150PX_VALUE = "150px";

    private static final String STYLE_WIDTH_16PX_VALUE = "16px";

    private static final String STYLE_WIDTH_50PX_VALUE = "50px";

    private transient List<ServerData> selectedServerDatas = null;

    private transient Map<ServerColumns, String> filterMap = null;

    private Checkbox enabledCheckbox = null;

    private Checkbox selectAllCheckbox = null;

    private DataProvider<ServerData, Void> dataProvider = null;

    private Grid<ServerData> serverGrid = null;

    private TextField nameTextField = null;

    private TextField noteTextField = null;

    public ServersGroupEditDialog( long serversGroupId, ServersGroupData serversGroupData, Database database, VirtualUtil virtualUtil )
    {
        filterMap = new EnumMap<>( ServerColumns.class );
        selectedServerDatas = serversGroupId == ServersGroupData.NEW_SERVERS_GROUP ? new ArrayList<>()
                        : database.getServersGroup( serversGroupId )
                                  .getServerDatas();
        // close/save buttons
        Button closeButton = new Button( virtualUtil.getTranslation( Translation.BUTTON_CLOSE ) );
        closeButton.getElement()
                   .setAttribute( ATTRIBUTE_THEME, ATTRIBUTE_THEME_VALUE );
        Button saveButton = new Button( virtualUtil.getTranslation( Translation.BUTTON_SAVE ) );
        saveButton.setEnabled( false );
        saveButton.addClassName( ATTRIBUTE_CLASS_BUTTON_VALUE );
        saveButton.getElement()
                  .setAttribute( ATTRIBUTE_THEME, ATTRIBUTE_THEME_VALUE );
        saveButton.getStyle()
                  .set( STYLE_MARGIN_LEFT, STYLE_MARGIN_VALUE );

        // buttons layout
        HorizontalLayout buttonsLayout = new HorizontalLayout( saveButton, closeButton );
        buttonsLayout.setJustifyContentMode( JustifyContentMode.CENTER );
        buttonsLayout.setMargin( true );
        buttonsLayout.setPadding( true );
        buttonsLayout.setSpacing( true );
        buttonsLayout.setWidth( STYLE_WIDTH_100PERCENT_VALUE );
        buttonsLayout.getStyle()
                     .set( STYLE_MARGIN_BOTTOM, STYLE_MARGIN_VALUE )
                     .set( STYLE_MARGIN_LEFT, STYLE_MARGIN_VALUE )
                     .set( STYLE_MARGIN_RIGHT, STYLE_MARGIN_VALUE )
                     .set( STYLE_PADDING_BOTTOM, STYLE_PADDING_VALUE )
                     .set( STYLE_PADDING_LEFT, STYLE_PADDING_VALUE )
                     .set( STYLE_PADDING_RIGHT, STYLE_PADDING_VALUE );

        // name
        nameTextField = new TextField( virtualUtil.getTranslation( Translation.FIELD_NAME ), StringUtils.stripToEmpty( serversGroupData.getName() ),
                                       event -> saveButton.setEnabled( isAllRequiredValuesNonEmpty() && !StringUtils.equals( event.getValue(), event.getOldValue() ) ) );
        nameTextField.setAutofocus( true );
        nameTextField.setRequiredIndicatorVisible( true );
        nameTextField.setValueChangeMode( ValueChangeMode.LAZY );
        nameTextField.setValueChangeTimeout( VALUE_CHANGE_TIMEOUT );
        nameTextField.getStyle()
                     .set( STYLE_PADDING_TOP, STYLE_PADDING_VALUE );

        // enabled
        enabledCheckbox = new Checkbox( virtualUtil.getTranslation( Translation.FIELD_ENABLED ),
                                        serversGroupId == ServersGroupData.NEW_SERVERS_GROUP ? ServersGroupData.DEFAULT_ENABLED : serversGroupData.isEnabled() );
        enabledCheckbox.addValueChangeListener( event -> saveButton.setEnabled( isAllRequiredValuesNonEmpty() && !Objects.equals( event.getValue(), event.getOldValue() ) ) );

        // name layout: name, enabled
        HorizontalLayout nameLayout = new HorizontalLayout( nameTextField, enabledCheckbox );
        nameLayout.expand( nameTextField );
        nameLayout.setDefaultVerticalComponentAlignment( Alignment.BASELINE );
        nameLayout.setPadding( false );
        nameLayout.setSpacing( true );
        nameLayout.setWidth( STYLE_WIDTH_100PERCENT_VALUE );

        // note
        noteTextField = new TextField( virtualUtil.getTranslation( Translation.FIELD_NOTE ), StringUtils.defaultIfBlank( serversGroupData.getNote(), StringUtils.EMPTY ),
                                       event -> saveButton.setEnabled( isAllRequiredValuesNonEmpty() && !StringUtils.equals( event.getValue(), event.getOldValue() ) ) );
        noteTextField.setValueChangeMode( ValueChangeMode.LAZY );
        noteTextField.setValueChangeTimeout( VALUE_CHANGE_TIMEOUT );
        noteTextField.setWidth( STYLE_WIDTH_100PERCENT_VALUE );
        noteTextField.getStyle()
                     .set( STYLE_PADDING_BOTTOM, STYLE_PADDING_14PX_VALUE );

        // server data provider
        dataProvider = DataProvider.fromCallbacks( query -> {
            int offset = query.getOffset();
            int limit = query.getLimit();

            List<String> orders = new ArrayList<>();
            for ( SortOrder<String> order : query.getSortOrders() )
            {
                orders.add( order.getDirection()
                                 .equals( SortDirection.DESCENDING ) ? Database.ORDER_DESCENDING + order.getSorted() : order.getSorted() );
            }

            return new ArrayList<>( database.getServers( offset, limit, filterMap, orders ) ).stream();
        }, query -> {
            long[] selectedServerIds = selectedServerDatas.stream()
                                                          .mapToLong( ServerData::getId )
                                                          .toArray();
            int filteredSelectedServersSize = //
                database.getServers( 0, NO_LIMITS, filterMap, null )
                        .stream()
                        .filter( serverData -> LongStream.of( selectedServerIds )
                                                         .anyMatch( selectedServerId -> selectedServerId == serverData.getId() ) )
                        .collect( Collectors.toList() )
                        .size();

            Long serverRows = database.getServerRows( filterMap );
            selectAllCheckbox.setValue( filteredSelectedServersSize == serverRows && serverRows > 0 );

            return serverRows.intValue();
        } );

        // server
        serverGrid = new Grid<>();
        serverGrid.setDataProvider( dataProvider );
        serverGrid.setId( CLASS_GRID_CELL_THEME );
        serverGrid.setPageSize( VISIBLE_ROWS );
        serverGrid.setSelectionMode( SelectionMode.NONE );
        serverGrid.getStyle()
                  .set( STYLE_DISPLAY, STYLE_DISPLAY_GRID_VALUE )
                  .set( STYLE_WIDTH, STYLE_WIDTH_100PERCENT_VALUE );

        // application: selected
        selectAllCheckbox = new Checkbox();
        selectAllCheckbox.addClickListener( event -> {
            selectedServerDatas.clear();
            if ( Boolean.TRUE.equals( event.getSource()
                                           .getValue() ) )
            {
                selectedServerDatas.addAll( database.getServers( 0, NO_LIMITS, filterMap, null ) );
            }
            saveButton.setEnabled( isAllRequiredValuesNonEmpty() );
            dataProvider.refreshAll();
        } );
        serverGrid.addColumn( new ComponentRenderer<>( serverData -> {
            Checkbox checkbox = new Checkbox();
            checkbox.setValue( selectedServerDatas.stream()
                                                  .anyMatch( selectedServerData -> Objects.equals( selectedServerData.getId(), serverData.getId() ) ) );
            checkbox.addClickListener( event -> {
                if ( Boolean.TRUE.equals( event.getSource()
                                               .getValue() ) )
                {
                    selectedServerDatas.add( serverData );
                }
                else
                {
                    selectedServerDatas.removeIf( selectedServerData -> Objects.equals( selectedServerData.getId(), serverData.getId() ) );
                }
                saveButton.setEnabled( isAllRequiredValuesNonEmpty() );
                dataProvider.refreshAll();
            } );
            return checkbox;
        } ) )
                  .setFlexGrow( 0 )
                  .setFrozen( true )
                  .setKey( ServerColumns.SELECTED.getName() )
                  .setResizable( false )
                  .setSortable( false )
                  .setWidth( STYLE_WIDTH_50PX_VALUE )
                  .setHeader( selectAllCheckbox );

        // server: name
        Grid.Column<ServerData> nameColumn = //
            serverGrid.addColumn( ServerData::getName )
                      .setFlexGrow( 1 )
                      .setFrozen( true )
                      .setKey( ServerColumns.NAME.getName() )
                      .setResizable( true )
                      .setSortable( true )
                      .setSortProperty( ServerColumns.NAME.getName() )
                      .setWidth( STYLE_WIDTH_150PX_VALUE )
                      .setHeader( virtualUtil.getTranslation( ServerColumns.NAME.getValueForTranslation() ) );

        // server: note
        Grid.Column<ServerData> noteColumn = //
            serverGrid.addColumn( ServerData::getNote )
                      .setFlexGrow( 2 )
                      .setKey( ServerColumns.NOTE.getName() )
                      .setResizable( true )
                      .setSortable( true )
                      .setSortProperty( ServerColumns.NOTE.getName() )
                      .setHeader( virtualUtil.getTranslation( ServerColumns.NOTE.getValueForTranslation() ) );

        // server: address
        Grid.Column<ServerData> addressColumn = //
            serverGrid.addColumn( ServerData::getAddress )
                      .setFlexGrow( 2 )
                      .setKey( ServerColumns.ADDRESS.getName() )
                      .setResizable( true )
                      .setSortable( true )
                      .setSortProperty( ServerColumns.ADDRESS.getName() )
                      .setWidth( STYLE_WIDTH_150PX_VALUE )
                      .setHeader( virtualUtil.getTranslation( ServerColumns.ADDRESS.getValueForTranslation() ) );

        // server: maximum users
        Grid.Column<ServerData> maxUsersColumn = //
            serverGrid.addColumn( ServerData::getMaxUsers )
                      .setFlexGrow( 0 )
                      .setKey( ServerColumns.MAX_USERS.getName() )
                      .setResizable( true )
                      .setSortable( true )
                      .setSortProperty( ServerColumns.MAX_USERS.getName() )
                      .setWidth( STYLE_WIDTH_130PX_VALUE )
                      .setHeader( virtualUtil.getTranslation( ServerColumns.MAX_USERS.getValueForTranslation() ) );

        // server: enabled
        Grid.Column<ServerData> enabledColumn = //
            serverGrid.addColumn( new ComponentRenderer<>( Div::new, ( div, serverData ) -> {
                Icon icon = serverData.isEnabled() ? new Icon( VaadinIcon.CHECK ) : new Icon( VaadinIcon.CLOSE );
                icon.getStyle()
                    .set( STYLE_HEIGHT, STYLE_HEIGHT_16PX_VALUE )
                    .set( STYLE_WIDTH, STYLE_WIDTH_16PX_VALUE );
                div.add( icon );
                div.getStyle()
                   .set( STYLE_TEXT_ALIGN, STYLE_TEXT_ALIGN_VALUE );
            } ) )
                      .setFlexGrow( 0 )
                      .setKey( ServerColumns.ENABLED.getName() )
                      .setResizable( true )
                      .setSortable( true )
                      .setSortProperty( ServerColumns.ENABLED.getName() )
                      .setWidth( STYLE_WIDTH_130PX_VALUE )
                      .setHeader( virtualUtil.getTranslation( ServerColumns.ENABLED.getValueForTranslation() ) );

        Label serverGridLabel = new Label( virtualUtil.getTranslation( Translation.SERVERS ) );
        serverGridLabel.setFor( virtualUtil.getTranslation( Translation.SERVERS ) );
        serverGridLabel.addClassName( CLASS_GRID_LABEL );

        VerticalLayout serverGridLayout = new VerticalLayout( serverGridLabel, serverGrid );
        serverGridLayout.setMargin( false );
        serverGridLayout.setPadding( false );
        serverGridLayout.setSpacing( false );
        serverGridLayout.setSizeUndefined();
        serverGridLayout.setWidth( STYLE_WIDTH_100PERCENT_VALUE );

        Binder<ServersGroupData> serversGroupBinder = new Binder<>();
        serversGroupBinder.forField( nameTextField )
                          .withValidator( new StringLengthValidator( virtualUtil.getTranslation( Translation.DATA_VALUE ), 1, null ) )
                          .bind( ServersGroupData::getName, ServersGroupData::setName );
        serversGroupBinder.bind( noteTextField, ServersGroupData::getNote, ServersGroupData::setNote );
        serversGroupBinder.bind( enabledCheckbox, ServersGroupData::isEnabled, ServersGroupData::setEnabled );

        closeButton.addClickListener( event -> close() );
        saveButton.addClickListener( event -> {
            ServersGroupData data = new ServersGroupData();
            if ( Objects.nonNull( data ) && serversGroupBinder.writeBeanIfValid( data ) )
            {
                data.setId( serversGroupId );
                data.setServerDatas( selectedServerDatas );
                savingData( data );
            }
            else
            {
                new DialogInfoUtil( virtualUtil.getTranslation( Translation.DATA_ERROR_SAVE ), virtualUtil.getTranslation( Translation.BUTTON_CLOSE ) ).open();
            }
        } );

        // filter: name
        HeaderRow filterRow = serverGrid.appendHeaderRow();
        TextField nameFilter = new TextField();
        nameFilter.setPlaceholder( virtualUtil.getTranslation( Translation.FIELD_FILTER ) );
        nameFilter.setSizeFull();
        nameFilter.setValueChangeMode( ValueChangeMode.LAZY );
        nameFilter.setValueChangeTimeout( VALUE_CHANGE_TIMEOUT );
        nameFilter.addValueChangeListener( event -> {
            if ( nameFilter.isEmpty() )
            {
                filterMap.remove( ServerColumns.NAME );
            }
            else
            {
                filterMap.put( ServerColumns.NAME, nameFilter.getValue() );
            }
            dataProvider.refreshAll();
        } );
        filterRow.getCell( nameColumn )
                 .setComponent( nameFilter );

        // filter: note
        TextField noteFilter = new TextField();
        noteFilter.setPlaceholder( virtualUtil.getTranslation( Translation.FIELD_FILTER ) );
        noteFilter.setSizeFull();
        noteFilter.setValueChangeMode( ValueChangeMode.LAZY );
        noteFilter.setValueChangeTimeout( VALUE_CHANGE_TIMEOUT );
        noteFilter.addValueChangeListener( event -> {
            if ( noteFilter.isEmpty() )
            {
                filterMap.remove( ServerColumns.NOTE );
            }
            else
            {
                filterMap.put( ServerColumns.NOTE, noteFilter.getValue() );
            }
            dataProvider.refreshAll();
        } );
        filterRow.getCell( noteColumn )
                 .setComponent( noteFilter );

        // filter: address
        TextField addressFilter = new TextField();
        addressFilter.setPlaceholder( virtualUtil.getTranslation( Translation.FIELD_FILTER ) );
        addressFilter.setSizeFull();
        addressFilter.setValueChangeMode( ValueChangeMode.LAZY );
        addressFilter.setValueChangeTimeout( VALUE_CHANGE_TIMEOUT );
        addressFilter.addValueChangeListener( event -> {
            if ( addressFilter.isEmpty() )
            {
                filterMap.remove( ServerColumns.ADDRESS );
            }
            else
            {
                filterMap.put( ServerColumns.ADDRESS, addressFilter.getValue() );
            }
            dataProvider.refreshAll();
        } );
        filterRow.getCell( addressColumn )
                 .setComponent( addressFilter );

        // filter: maximum users
        TextField maxUsersFilter = new TextField();
        maxUsersFilter.setPlaceholder( virtualUtil.getTranslation( Translation.FIELD_FILTER ) );
        maxUsersFilter.setSizeFull();
        maxUsersFilter.setValueChangeMode( ValueChangeMode.LAZY );
        maxUsersFilter.setValueChangeTimeout( VALUE_CHANGE_TIMEOUT );
        maxUsersFilter.addValueChangeListener( event -> {
            if ( maxUsersFilter.isEmpty() )
            {
                filterMap.remove( ServerColumns.MAX_USERS );
            }
            else
            {
                filterMap.put( ServerColumns.MAX_USERS, maxUsersFilter.getValue() );
            }
            dataProvider.refreshAll();
        } );
        filterRow.getCell( maxUsersColumn )
                 .setComponent( maxUsersFilter );

        // filter: enabled
        String filterAll = virtualUtil.getTranslation( Translation.DATA_FILTER_ALL );
        String filterNo = virtualUtil.getTranslation( Translation.DATA_FILTER_NO );
        String filterYes = virtualUtil.getTranslation( Translation.DATA_FILTER_YES );
        ComboBox<String> enabledFilter = new ComboBox<>();
        enabledFilter.setAllowCustomValue( false );
        enabledFilter.setClassName( STYLE_GRID_COMBOBOX );
        enabledFilter.setItems( filterAll, filterYes, filterNo );
        enabledFilter.setValue( filterAll );
        enabledFilter.setSizeFull();
        enabledFilter.addValueChangeListener( event -> {
            if ( StringUtils.equals( enabledFilter.getValue(), filterYes ) )
            {
                filterMap.put( ServerColumns.ENABLED, Boolean.TRUE.toString() );
            }
            else if ( StringUtils.equals( enabledFilter.getValue(), filterNo ) )
            {
                filterMap.put( ServerColumns.ENABLED, Boolean.FALSE.toString() );
            }
            else
            {
                filterMap.remove( ServerColumns.ENABLED );
            }
            dataProvider.refreshAll();
        } );
        filterRow.getCell( enabledColumn )
                 .setComponent( enabledFilter );

        // form layout
        setFormLayout( new FormLayout( nameLayout, noteTextField, serverGridLayout, buttonsLayout ) );
        getFormLayout().setClassName( CLASS_DIALOG );
        getFormLayout().setResponsiveSteps( new ResponsiveStep( FORM_MINIMAL_WIDTH, FORM_COLUMNS, LabelsPosition.TOP ) );
        getFormLayout().setWidth( STYLE_WIDTH_100PERCENT_VALUE );
        add( getFormLayout() );

        // set active/selected rows
        serverGrid.getColumnByKey( ServerColumns.ADDRESS.getName() )
                  .setClassNameGenerator( serverData -> getCellClassName( serverData.getId() ) );
        serverGrid.getColumnByKey( ServerColumns.ENABLED.getName() )
                  .setClassNameGenerator( serverData -> getCellClassName( serverData.getId() ) );
        serverGrid.getColumnByKey( ServerColumns.MAX_USERS.getName() )
                  .setClassNameGenerator( serverData -> getCellClassName( serverData.getId() ) );
        serverGrid.getColumnByKey( ServerColumns.NAME.getName() )
                  .setClassNameGenerator( serverData -> getCellClassName( serverData.getId() ) );
        serverGrid.getColumnByKey( ServerColumns.NOTE.getName() )
                  .setClassNameGenerator( serverData -> getCellClassName( serverData.getId() ) );
        serverGrid.getColumnByKey( ServerColumns.SELECTED.getName() )
                  .setClassNameGenerator( serverData -> getCellClassName( serverData.getId() ) );
    }

    public abstract void savingData( ServersGroupData serversGroupData );

    private boolean isAllRequiredValuesNonEmpty()
    {
        return ( StringUtils.isNotBlank( nameTextField.getValue() ) && CollectionUtils.isNotEmpty( selectedServerDatas ) );
    }

    private String getCellClassName( Long sevrerDataId )
    {
        if ( selectedServerDatas.stream()
                                .anyMatch( selectedServerData -> Objects.equals( selectedServerData.getId(), sevrerDataId ) ) )
        {
            return CLASS_GRID_CELL_ACTIVE;
        }
        return CLASS_GRID_CELL_NOT_ACTIVE;
    }

}
