/**
 * Copyright (C) 2020-2022 Dušan Vejnović <vaadin@dussan.org>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.dussan.vaadin.dvirtual.utils;

import java.io.Serializable;
import java.security.InvalidAlgorithmParameterException;
import java.security.InvalidKeyException;
import java.security.NoSuchAlgorithmException;
import java.security.spec.InvalidKeySpecException;
import java.security.spec.KeySpec;
import java.util.Base64;
import java.util.Objects;

import javax.crypto.BadPaddingException;
import javax.crypto.Cipher;
import javax.crypto.IllegalBlockSizeException;
import javax.crypto.NoSuchPaddingException;
import javax.crypto.SecretKey;
import javax.crypto.SecretKeyFactory;
import javax.crypto.spec.GCMParameterSpec;
import javax.crypto.spec.PBEKeySpec;
import javax.crypto.spec.SecretKeySpec;

import org.apache.commons.lang3.StringUtils;
import org.slf4j.Logger;

public class EncryptionUtil
    implements Serializable
{

    private static final long serialVersionUID = 2532266829412328750L;

    private static final Logger LOGGER = LoggerUtil.getLogger();

    private static final int IV_LENGTH = 12;

    private static final int SALT_KEY_LENGTH = 32;

    private static final int SALT_LENGTH = 32;

    private static final int SECRET_KEY_ITERACTION_COUNT = 65536;

    private static final int SECRET_KEY_LENGTH = 256;

    private static final int TAG_LENGTH = 16;

    private static final String CIPHER_INSTANCE = "AES/GCM/NoPadding";

    private static final String ENCRYPT_ERROR = "Error while encrypting/decrypting!";

    private static final String IV_STRING = "POs4HuexT6wa";

    private static final String SALT_KEY_STRING = "XqBwYSrpANODyJILVPzRClyspNpseV2j";

    private static final String SALT_STRING = "DeIMDempGvCBy2FMLlLfd659gx8FyAWQ";

    private static final String SECRET_KEY_INSTANCE = "PBKDF2WithHmacSHA256";

    private static final String SECRET_KEY_SPEC = "AES";

    public static SecretKeySpec getSecretKeySpecSpec( String salt, String saltKey )
    {
        final SecretKeySpec secretKeySpec;
        try
        {
            SecretKeyFactory secretKeyFactory = SecretKeyFactory.getInstance( SECRET_KEY_INSTANCE );
            KeySpec keySpec = new PBEKeySpec( StringUtils.left( saltKey + SALT_KEY_STRING, SALT_KEY_LENGTH )
                                                         .toCharArray(),
                                              StringUtils.left( salt + SALT_STRING, SALT_LENGTH )
                                                         .getBytes(),
                                              SECRET_KEY_ITERACTION_COUNT, SECRET_KEY_LENGTH );
            SecretKey secretKey = secretKeyFactory.generateSecret( keySpec );
            secretKeySpec = new SecretKeySpec( secretKey.getEncoded(), SECRET_KEY_SPEC );
        }
        catch ( NoSuchAlgorithmException | InvalidKeySpecException exception )
        {
            return null;
        }
        return secretKeySpec;
    }

    public static byte[] getSecureIV( String saltIV )
    {
        return StringUtils.left( saltIV + IV_STRING, IV_LENGTH )
                          .getBytes();
    }

    public static String encrypt( String value, SecretKeySpec secretKeySpec, byte[] secureIV )
    {
        byte[] encryptedText = null;
        try
        {
            Cipher cipher = Cipher.getInstance( CIPHER_INSTANCE );
            GCMParameterSpec gcmParameterSpec = new GCMParameterSpec( TAG_LENGTH * 8, secureIV );
            cipher.init( Cipher.ENCRYPT_MODE, secretKeySpec, gcmParameterSpec );
            encryptedText = cipher.doFinal( value.getBytes() );
        }
        catch ( BadPaddingException | InvalidAlgorithmParameterException | IllegalBlockSizeException | InvalidKeyException | NullPointerException | NoSuchAlgorithmException
                        | NoSuchPaddingException exception )
        {
            LOGGER.warn( ENCRYPT_ERROR );
        }
        return Base64.getEncoder()
                     .encodeToString( encryptedText );
    }

    public static String decrypt( String value, SecretKeySpec secretKeySpec, byte[] secureIV )
    {
        byte[] decryptedText = null;
        try
        {
            Cipher cipher = Cipher.getInstance( CIPHER_INSTANCE );
            GCMParameterSpec gcmParameterSpec = new GCMParameterSpec( TAG_LENGTH * 8, secureIV );
            cipher.init( Cipher.DECRYPT_MODE, secretKeySpec, gcmParameterSpec );
            decryptedText = cipher.doFinal( Base64.getDecoder()
                                                  .decode( value ) );
        }
        catch ( BadPaddingException | InvalidAlgorithmParameterException | IllegalBlockSizeException | InvalidKeyException | NullPointerException | NoSuchAlgorithmException
                        | NoSuchPaddingException exception )
        {
            LOGGER.warn( ENCRYPT_ERROR );
        }
        return ( Objects.nonNull( decryptedText ) ? new String( decryptedText ) : StringUtils.EMPTY );
    }

}
