/**
 * Copyright (C) 2020-2022 Dušan Vejnović <vaadin@dussan.org>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.dussan.vaadin.dvirtual.utils;

import java.io.FileReader;
import java.io.IOException;
import java.io.Serializable;
import java.nio.file.Paths;
import java.util.Objects;
import java.util.Properties;

import org.dussan.vaadin.dvirtual.database.data.ApplicationData;
import org.dussan.vaadin.dvirtual.database.data.ApplicationServersGroupData;
import org.dussan.vaadin.dvirtual.database.data.ApplicationsGroupApplicationData;
import org.dussan.vaadin.dvirtual.database.data.ApplicationsGroupData;
import org.dussan.vaadin.dvirtual.database.data.IconData;
import org.dussan.vaadin.dvirtual.database.data.ServerData;
import org.dussan.vaadin.dvirtual.database.data.ServersGroupData;
import org.dussan.vaadin.dvirtual.database.data.ServersGroupServerData;
import org.dussan.vaadin.dvirtual.database.data.UserApplicationData;
import org.dussan.vaadin.dvirtual.database.data.UserApplicationsGroupData;
import org.dussan.vaadin.dvirtual.database.data.UserData;
import org.dussan.vaadin.dvirtual.database.data.UserSessionApplicationData;
import org.dussan.vaadin.dvirtual.database.data.UserSessionData;
import org.hibernate.SessionFactory;
import org.hibernate.boot.Metadata;
import org.hibernate.boot.MetadataBuilder;
import org.hibernate.boot.MetadataSources;
import org.hibernate.boot.SessionFactoryBuilder;
import org.hibernate.boot.registry.StandardServiceRegistryBuilder;
import org.hibernate.service.ServiceRegistry;
import org.hibernate.service.spi.ServiceException;
import org.slf4j.Logger;

public class HibernateUtil
    implements Serializable
{

    private static final long serialVersionUID = 6448848617637360336L;

    private static final Logger LOGGER = LoggerUtil.getLogger();

    private static final String ERROR_CANNOT_LOAD = "Cannot load database configuration, cause: {}";

    private static final String ERROR_CANNOT_START = "Cannot start database service, cause: {}";

    private String configFile = null;

    private String configFolder = null;

    private SessionFactory sessionFactory = null;

    public HibernateUtil( String configFolder, String configFile )
    {
        this.configFolder = configFolder;
        this.configFile = configFile;
    }

    private Properties getProperties()
    {
        Properties properties = new Properties();
        try (FileReader fileReader = new FileReader( Paths.get( configFolder, configFile )
                                                          .toFile() ))
        {
            properties.load( fileReader );
        }
        catch ( IOException exception )
        {
            LOGGER.error( ERROR_CANNOT_LOAD, exception.getMessage() );
        }
        return properties;
    }

    private static ServiceRegistry configureServiceRegistry( Properties properties )
    {
        StandardServiceRegistryBuilder standardServiceRegistryBuilder = new StandardServiceRegistryBuilder().applySettings( properties );
        return standardServiceRegistryBuilder.build();
    }

    private static SessionFactory getSessionFactory( ServiceRegistry serviceRegistry )
    {
        try
        {
            MetadataSources metadataSources = new MetadataSources( serviceRegistry );
            metadataSources.addAnnotatedClass( ApplicationData.class );
            metadataSources.addAnnotatedClass( ApplicationServersGroupData.class );
            metadataSources.addAnnotatedClass( ApplicationsGroupApplicationData.class );
            metadataSources.addAnnotatedClass( ApplicationsGroupData.class );
            metadataSources.addAnnotatedClass( IconData.class );
            metadataSources.addAnnotatedClass( ServerData.class );
            metadataSources.addAnnotatedClass( ServersGroupData.class );
            metadataSources.addAnnotatedClass( ServersGroupServerData.class );
            metadataSources.addAnnotatedClass( UserApplicationData.class );
            metadataSources.addAnnotatedClass( UserApplicationsGroupData.class );
            metadataSources.addAnnotatedClass( UserData.class );
            metadataSources.addAnnotatedClass( UserSessionApplicationData.class );
            metadataSources.addAnnotatedClass( UserSessionData.class );

            MetadataBuilder metadataBuilder = metadataSources.getMetadataBuilder();
            Metadata metadata = metadataBuilder.build();
            SessionFactoryBuilder sessionFactoryBuilder = metadata.getSessionFactoryBuilder();
            return sessionFactoryBuilder.build();
        }
        catch ( ServiceException exception )
        {
            LOGGER.error( ERROR_CANNOT_START, exception.getMessage() );
        }
        return null;
    }

    public SessionFactory getSessionFactory()
    {
        if ( Objects.isNull( sessionFactory ) || sessionFactory.isClosed() )
        {
            Properties properties = getProperties();
            ServiceRegistry serviceRegistry = configureServiceRegistry( properties );
            sessionFactory = getSessionFactory( serviceRegistry );
        }
        return sessionFactory;
    }

    public boolean isInitialized()
    {
        return Objects.nonNull( getSessionFactory() );
    }

}
