/**
 * Copyright (C) 2020-2022 Dušan Vejnović <vaadin@dussan.org>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.dussan.vaadin.dvirtual.xpra.client;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.regex.Pattern;

import org.apache.commons.collections4.MapUtils;
import org.apache.commons.lang3.ArrayUtils;
import org.apache.commons.lang3.StringUtils;
import org.dussan.vaadin.dvirtual.client.VirtualClient.VirtualClientAction;
import org.dussan.vaadin.dvirtual.info.KeyboardKeyCodes;
import org.dussan.vaadin.dvirtual.info.ClientProperties;
import org.dussan.vaadin.dvirtual.utils.CanvasImageUtil;
import org.dussan.vaadin.dvirtual.utils.VirtualUtil;
import org.dussan.vaadin.dvirtual.utils.LoggerUtil;
import org.dussan.vaadin.dvirtual.xpra.info.FullCscMode;
import org.dussan.vaadin.dvirtual.xpra.info.PictureEncodingType;
import org.dussan.vaadin.dvirtual.xpra.info.RgbFormat;
import org.dussan.vaadin.dvirtual.xpra.io.Receiver;
import org.dussan.vaadin.dvirtual.xpra.io.Sender;
import org.dussan.vaadin.dvirtual.xpra.packets.PacketType;
import org.dussan.vaadin.dvirtual.xpra.packets.base.ClipboardTarget;
import org.dussan.vaadin.dvirtual.xpra.packets.base.WindowType;
import org.dussan.vaadin.dvirtual.xpra.packets.client.ClientButtonActionPacket;
import org.dussan.vaadin.dvirtual.xpra.packets.client.ClientClipboardContentsNonePacket;
import org.dussan.vaadin.dvirtual.xpra.packets.client.ClientClipboardContentsPacket;
import org.dussan.vaadin.dvirtual.xpra.packets.client.ClientClipboardTokenPacket;
import org.dussan.vaadin.dvirtual.xpra.packets.client.ClientCloseWindowPacket;
import org.dussan.vaadin.dvirtual.xpra.packets.client.ClientConfigureWindowPacket;
import org.dussan.vaadin.dvirtual.xpra.packets.client.ClientDamageSequencePacket;
import org.dussan.vaadin.dvirtual.xpra.packets.client.ClientDesktopSizePacket;
import org.dussan.vaadin.dvirtual.xpra.packets.client.ClientDisconnectPacket;
import org.dussan.vaadin.dvirtual.xpra.packets.client.ClientFocusPacket;
import org.dussan.vaadin.dvirtual.xpra.packets.client.ClientHelloPacket;
import org.dussan.vaadin.dvirtual.xpra.packets.client.ClientKeyActionPacket;
import org.dussan.vaadin.dvirtual.xpra.packets.client.ClientMapWindowPacket;
import org.dussan.vaadin.dvirtual.xpra.packets.client.ClientPingEchoPacket;
import org.dussan.vaadin.dvirtual.xpra.packets.client.ClientPingPacket;
import org.dussan.vaadin.dvirtual.xpra.packets.client.ClientPointerPositionPacket;
import org.dussan.vaadin.dvirtual.xpra.packets.client.ClientResumePacket;
import org.dussan.vaadin.dvirtual.xpra.packets.client.ClientSetDeflatePacket;
import org.dussan.vaadin.dvirtual.xpra.packets.client.ClientSuspendPacket;
import org.dussan.vaadin.dvirtual.xpra.packets.client.ClientUnmapWindowPacket;
import org.dussan.vaadin.dvirtual.xpra.packets.client.ClientWheelMotionPacket;
import org.dussan.vaadin.dvirtual.xpra.packets.server.ServerClipboardRequestPacket;
import org.dussan.vaadin.dvirtual.xpra.packets.server.ServerClipboardTokenPacket;
import org.dussan.vaadin.dvirtual.xpra.packets.server.ServerConfigureOverrideRedirectPacket;
import org.dussan.vaadin.dvirtual.xpra.packets.server.ServerCursorPacket;
import org.dussan.vaadin.dvirtual.xpra.packets.server.ServerDesktopSizePacket;
import org.dussan.vaadin.dvirtual.xpra.packets.server.ServerDisconnectPacket;
import org.dussan.vaadin.dvirtual.xpra.packets.server.ServerDrawPacket;
import org.dussan.vaadin.dvirtual.xpra.packets.server.ServerEncodingsPacket;
import org.dussan.vaadin.dvirtual.xpra.packets.server.ServerEosPacket;
import org.dussan.vaadin.dvirtual.xpra.packets.server.ServerHelloPacket;
import org.dussan.vaadin.dvirtual.xpra.packets.server.ServerLostWindowPacket;
import org.dussan.vaadin.dvirtual.xpra.packets.server.ServerNewOverrideRedirectPacket;
import org.dussan.vaadin.dvirtual.xpra.packets.server.ServerNewWindowPacket;
import org.dussan.vaadin.dvirtual.xpra.packets.server.ServerPingEchoPacket;
import org.dussan.vaadin.dvirtual.xpra.packets.server.ServerPingPacket;
import org.dussan.vaadin.dvirtual.xpra.packets.server.ServerRaiseWindowPacket;
import org.dussan.vaadin.dvirtual.xpra.packets.server.ServerSetClipboardEnabledPacket;
import org.dussan.vaadin.dvirtual.xpra.packets.server.ServerSetDeflatePacket;
import org.dussan.vaadin.dvirtual.xpra.packets.server.ServerStartupCompletePacket;
import org.dussan.vaadin.dvirtual.xpra.packets.server.ServerWindowIconPacket;
import org.dussan.vaadin.dvirtual.xpra.packets.server.ServerWindowMetadataPacket;
import org.dussan.vaadin.dvirtual.xpra.packets.server.ServerWindowMoveResizePacket;
import org.dussan.vaadin.dvirtual.xpra.packets.server.ServerWindowResizedPacket;
import org.dussan.vaadin.dvirtual.xpra.processors.client.ClientButtonActionPacketProcessor;
import org.dussan.vaadin.dvirtual.xpra.processors.client.ClientClipboardContentsNonePacketProcessor;
import org.dussan.vaadin.dvirtual.xpra.processors.client.ClientClipboardContentsPacketProcessor;
import org.dussan.vaadin.dvirtual.xpra.processors.client.ClientClipboardTokenPacketProcessor;
import org.dussan.vaadin.dvirtual.xpra.processors.client.ClientCloseWindowPacketProcessor;
import org.dussan.vaadin.dvirtual.xpra.processors.client.ClientConfigureWindowPacketProcessor;
import org.dussan.vaadin.dvirtual.xpra.processors.client.ClientDamageSequencePacketProcessor;
import org.dussan.vaadin.dvirtual.xpra.processors.client.ClientDesktopSizePacketProcessor;
import org.dussan.vaadin.dvirtual.xpra.processors.client.ClientDisconnectPacketProcessor;
import org.dussan.vaadin.dvirtual.xpra.processors.client.ClientFocusPacketProcessor;
import org.dussan.vaadin.dvirtual.xpra.processors.client.ClientHelloPacketProcessor;
import org.dussan.vaadin.dvirtual.xpra.processors.client.ClientKeyActionPacketProcessor;
import org.dussan.vaadin.dvirtual.xpra.processors.client.ClientMapWindowPacketProcessor;
import org.dussan.vaadin.dvirtual.xpra.processors.client.ClientPingEchoPacketProcessor;
import org.dussan.vaadin.dvirtual.xpra.processors.client.ClientPingPacketProcessor;
import org.dussan.vaadin.dvirtual.xpra.processors.client.ClientPointerPositionPacketProcessor;
import org.dussan.vaadin.dvirtual.xpra.processors.client.ClientResumePacketProcessor;
import org.dussan.vaadin.dvirtual.xpra.processors.client.ClientSetDeflatePacketProcessor;
import org.dussan.vaadin.dvirtual.xpra.processors.client.ClientSuspendPacketProcessor;
import org.dussan.vaadin.dvirtual.xpra.processors.client.ClientUnmapWindowPacketProcessor;
import org.dussan.vaadin.dvirtual.xpra.processors.client.ClientWheelMotionPacketProcessor;
import org.dussan.vaadin.dvirtual.xpra.processors.server.ServerClipboardRequestPacketProcessor;
import org.dussan.vaadin.dvirtual.xpra.processors.server.ServerClipboardTokenPacketProcessor;
import org.dussan.vaadin.dvirtual.xpra.processors.server.ServerConfigureOverrideRedirectPacketProcessor;
import org.dussan.vaadin.dvirtual.xpra.processors.server.ServerCursorPacketProcessor;
import org.dussan.vaadin.dvirtual.xpra.processors.server.ServerDesktopSizePacketProcessor;
import org.dussan.vaadin.dvirtual.xpra.processors.server.ServerDisconnectPacketProcessor;
import org.dussan.vaadin.dvirtual.xpra.processors.server.ServerDrawPacketProcessor;
import org.dussan.vaadin.dvirtual.xpra.processors.server.ServerEncodingsPacketProcessor;
import org.dussan.vaadin.dvirtual.xpra.processors.server.ServerEosPacketProcessor;
import org.dussan.vaadin.dvirtual.xpra.processors.server.ServerHelloPacketProcessor;
import org.dussan.vaadin.dvirtual.xpra.processors.server.ServerLostWindowPacketProcessor;
import org.dussan.vaadin.dvirtual.xpra.processors.server.ServerNewOverrideRedirectPacketProcessor;
import org.dussan.vaadin.dvirtual.xpra.processors.server.ServerNewWindowPacketProcessor;
import org.dussan.vaadin.dvirtual.xpra.processors.server.ServerPingEchoPacketProcessor;
import org.dussan.vaadin.dvirtual.xpra.processors.server.ServerPingPacketProcessor;
import org.dussan.vaadin.dvirtual.xpra.processors.server.ServerRaiseWindowPacketProcessor;
import org.dussan.vaadin.dvirtual.xpra.processors.server.ServerSetClipboardEnabledPacketProcessor;
import org.dussan.vaadin.dvirtual.xpra.processors.server.ServerSetDeflatePacketProcessor;
import org.dussan.vaadin.dvirtual.xpra.processors.server.ServerStartupCompletePacketProcessor;
import org.dussan.vaadin.dvirtual.xpra.processors.server.ServerWindowIconPacketProcessor;
import org.dussan.vaadin.dvirtual.xpra.processors.server.ServerWindowMetadataPacketProcessor;
import org.dussan.vaadin.dvirtual.xpra.processors.server.ServerWindowMoveResizePacketProcessor;
import org.dussan.vaadin.dvirtual.xpra.processors.server.ServerWindowResizedPacketProcessor;
import org.dussan.vaadin.dvirtual.xpra.window.Cursor;
import org.dussan.vaadin.dvirtual.xpra.window.Window;
import org.dussan.vaadin.dvirtual.xpra.window.WindowDimension;
import org.slf4j.Logger;

public abstract class Client
    implements Serializable
{

    private static final long serialVersionUID = -7377554760044715671L;

    private static final Logger LOGGER = LoggerUtil.getLogger();

    private static final int NO_PARENT_WINDOW = 0;

    private static final int ONLY_ONE_WINDOW = 1;

    private static final int UNDEFINED_WINDOW = -1;

    private static final int WINDOW_MIN_HEIGHT = 100;

    private static final int WINDOW_MIN_WIDTH = 200;

    private static final int WINDOW_UNVISIBLE_HEIGHT = 10;

    private static final int WINDOW_UNVISIBLE_WIDTH = 10;

    private static final int WINDOW_UNVISIBLE_X = -10000;

    private static final int WINDOW_UNVISIBLE_Y = -10000;

    private static final String CANNOT_DRAW_INTO_UNDEFINED_WINDOW = "Cannot draw into the undefined window!";

    private static final String CANNOT_DRAW_NO_DATA = "Cannot draw because no data!";

    private static final String CANNOT_DRAW_UNSUPPORTED_COMPRESSION_LIBRARY = "Cannot draw because of the unsupported compression library!";

    private static final String CANNOT_DRAW_UNSUPPORTED_ENCODING_TYPE = "Cannot draw because of the unsupported encoding type: {}!";

    private static final String CONNECTION_INTERRUPTED = "Connection to the server is interrupted by the user.";

    private static final String CONNECTION_INTERRUPTED_BY_CLIENT_REQUEST = "client request";

    private static final String CONNECTION_INTERRUPTED_BY_SERVER = "Connection from the server is interrupted by server, cause: {}.";

    private static final String CONNECTION_INTERRUPTED_BY_SERVER_DETACH = "detach request";

    private static final String CONNECTION_INTERRUPTED_BY_SERVER_ERROR = "server error";

    private static final String CONNECTION_INTERRUPTED_BY_SERVER_SHUTDOWN = "server shutdown";

    private static final String CONNECTION_INTERRUPTED_BY_USER = "Connection to the server is interrupted by the user, cause: {}.";

    private static final String DATA_ARRAY = "_::_";

    private static final String DATA_ELEMENT = "_:_";

    private static final String DATA_TYPE_LZ4 = "lz4";

    private static final String DATA_TYPE_LZ4_RGB32 = DATA_TYPE_LZ4 + "_" + PictureEncodingType.RGB32.toString();

    private static final String DATA_TYPE_RGB = PictureEncodingType.RGB.toString();

    private static final String DATA_TYPE_RGB32 = PictureEncodingType.RGB32.toString();

    private static final String DATA_TYPE_ZLIB = "zlib";

    private static final String DATA_TYPE_ZLIB_RGB = DATA_TYPE_ZLIB + "_" + PictureEncodingType.RGB.toString();

    private static final String DATA_TYPE_ZLIB_RGB32 = DATA_TYPE_ZLIB + "_" + PictureEncodingType.RGB32.toString();

    private static final String MAXIMIZED_NO_INDICATOR = "-%s;";

    private static final String MAXIMIZED_YES_INDICATOR = "+%s;";

    private static final String PATTERN_WHITE_SPACES = "[\\s]";

    private static final String PATTERN_WHITE_SPACES_REPLACEMENT = StringUtils.EMPTY;

    private static final String STARTUP_COMPLETE = "Startup complete.";

    private static final String WINDOW_EXISTS = "Window with id {} already exists.";

    private boolean applicationRunning = true;

    private boolean clipboardEnabled = true;

    private boolean disconnectedByClientRequest = false;

    private boolean disconnectedBySecurityError = false;

    private boolean disconnectedByServer = false;

    private boolean disconnectedByServerDetach = false;

    private boolean disconnectedByServerError = false;

    private boolean disconnectedByServerShutdown = false;

    private boolean forceUpdateEnabled = false;

    private boolean mousePreciseWheel = false;

    private boolean updateEnabled = false;

    private boolean windowAutoCenterEnabled = true;

    private boolean windowCreated = false;

    private boolean windowFocusForced = false;

    private int compressionLevel = 0;

    private int createdWindowId = 0;

    private int desktopHeight = 0;

    private int desktopWidth = 0;

    private int dpi = 0;

    private int drawingDataWindowId = 0;

    private int focusedWindowId = 0;

    private String clipboardData = null;

    private String colorGamut = null;

    private String keyboardLayout = null;

    private transient List<List<Object>> keyCodes = null;

    private transient List<String> modifiers = null;

    private transient Map<Integer, Window> windows = null;

    private Pattern pattern = null;

    private StringBuilder drawingData = null;

    private Cursor cursor = null;

    private VirtualUtil virtualUtil = null;

    private FullCscMode[] fullCscModes = null;

    private PictureEncodingType[] pictureEncodingTypes = null;

    private Receiver receiver = null;

    private RgbFormat[] rgbFormats = null;

    private Sender sender = null;

    private Client()
    {
        windows = new HashMap<>();
        createdWindowId = UNDEFINED_WINDOW;
        focusedWindowId = UNDEFINED_WINDOW;
        desktopHeight = ClientProperties.DEFAULT_HEIGHT;
        desktopWidth = ClientProperties.DEFAULT_WIDTH;
        dpi = ClientProperties.DPI;
        pattern = Pattern.compile( PATTERN_WHITE_SPACES );
        drawingData = new StringBuilder();
    }

    protected Client( ClientProperties clientProperties, VirtualUtil virtualUtil )
    {
        this();
        this.desktopWidth = clientProperties.getWidth();
        this.desktopHeight = clientProperties.getHeight();
        this.dpi = clientProperties.getDpi() > 0 ? clientProperties.getDpi() : this.dpi;
        this.colorGamut = clientProperties.getColorGamut();
        this.keyboardLayout = virtualUtil.getKeyboardLayout()
                                         .getLayoutCode();
        this.keyCodes = KeyboardKeyCodes.getKeyCodes();
        this.pictureEncodingTypes = PictureEncodingType.getAllTypes( clientProperties.isWebpEnabled() );
        this.rgbFormats = RgbFormat.getAllFormats();
        this.fullCscModes = FullCscMode.getAllCscModes();
        this.virtualUtil = virtualUtil;

        receiver = new Receiver();
        activateServerPacketProcessors();
    }

    public abstract void virtualClientAction( Integer id, VirtualClientAction action, Object... data );

    private String removeWhiteSpacesFromString( String stringData )
    {
        return pattern.matcher( stringData )
                      .replaceAll( PATTERN_WHITE_SPACES_REPLACEMENT );
    }

    private String getDrawingData()
    {
        return removeWhiteSpacesFromString( drawingData.toString() );
    }

    private void addDrawingData( ServerDrawPacket packet )
    {
        if ( packet.getImage().length > 0 )
        {
            // add image position, image dimensions and image data
            drawingData.append( packet.getX() );
            drawingData.append( DATA_ELEMENT );
            drawingData.append( packet.getY() );
            drawingData.append( DATA_ELEMENT );
            drawingData.append( packet.getWidth() );
            drawingData.append( DATA_ELEMENT );
            drawingData.append( packet.getHeight() );
            drawingData.append( DATA_ELEMENT );
            drawingData.append( packet.getRowStride() );
            drawingData.append( DATA_ELEMENT );
            drawingData.append( packet.getPacketSequence() );
            drawingData.append( DATA_ELEMENT );

            // add image coding type
            switch ( packet.getPictureEncodingType() )
            {
                case H264:
                case SCROLL:
                    drawingData.append( CanvasImageUtil.toUint8Array( packet.getImage() ) );
                    drawingData.append( DATA_ELEMENT );
                    drawingData.append( packet.getPictureEncodingType()
                                              .toString() );
                    break;
                case JPEG:
                case PNG:
                case RGB:
                case WEBP:
                    drawingData.append( CanvasImageUtil.toBase64( packet.getImage() ) );
                    drawingData.append( DATA_ELEMENT );
                    drawingData.append( packet.getPictureEncodingType()
                                              .toString() );
                    break;
                case RGB24:
                    if ( MapUtils.isNotEmpty( packet.getOptions() ) && Objects.nonNull( MapUtils.getObject( packet.getOptions(), DATA_TYPE_ZLIB ) ) )
                    {
                        drawingData.append( CanvasImageUtil.toUint8Array( packet.getImage() ) );
                        drawingData.append( DATA_ELEMENT );
                        drawingData.append( DATA_TYPE_ZLIB_RGB );
                    }
                    else
                    {
                        drawingData.append( CanvasImageUtil.toUint8Array( packet.getImage() ) );
                        drawingData.append( DATA_ELEMENT );
                        drawingData.append( DATA_TYPE_RGB );
                    }
                    break;
                case RGB32:
                    if ( MapUtils.isNotEmpty( packet.getOptions() )
                        && ( Objects.nonNull( MapUtils.getObject( packet.getOptions(), DATA_TYPE_LZ4 ) ) || Objects.nonNull( MapUtils.getObject( packet.getOptions(), DATA_TYPE_ZLIB ) ) ) )
                    {
                        if ( Objects.nonNull( MapUtils.getObject( packet.getOptions(), DATA_TYPE_LZ4 ) ) )
                        {
                            drawingData.append( CanvasImageUtil.toUint8Array( packet.getImage() ) );
                            drawingData.append( DATA_ELEMENT );
                            drawingData.append( DATA_TYPE_LZ4_RGB32 );
                        }
                        else if ( Objects.nonNull( MapUtils.getObject( packet.getOptions(), DATA_TYPE_ZLIB ) ) )
                        {
                            drawingData.append( CanvasImageUtil.toUint8Array( CanvasImageUtil.uncompress( packet.getImage() ) ) );
                            drawingData.append( DATA_ELEMENT );
                            drawingData.append( DATA_TYPE_ZLIB_RGB32 );
                        }
                        else
                        {
                            LOGGER.warn( CANNOT_DRAW_UNSUPPORTED_COMPRESSION_LIBRARY );
                        }
                    }
                    else
                    {
                        drawingData.append( CanvasImageUtil.toUint8Array( packet.getImage() ) );
                        drawingData.append( DATA_ELEMENT );
                        drawingData.append( DATA_TYPE_RGB32 );
                    }
                    break;
                case UNDEFINED:
                default:
                    PictureEncodingType pictureEncodingType = packet.getPictureEncodingType();
                    String unsupportedEncodingType = pictureEncodingType.getUndefinedType();
                    unsupportedEncodingType = StringUtils.defaultIfBlank( unsupportedEncodingType, pictureEncodingType.toString() );
                    drawingData.append( CanvasImageUtil.toUint8Array( packet.getImage() ) );
                    drawingData.append( DATA_ELEMENT );
                    drawingData.append( unsupportedEncodingType );
                    LOGGER.warn( CANNOT_DRAW_UNSUPPORTED_ENCODING_TYPE, unsupportedEncodingType );
                    break;
            }

            // add array splitter
            drawingData.append( DATA_ARRAY );
        }
        else
        {
            LOGGER.error( CANNOT_DRAW_NO_DATA );
        }
    }

    private void resetDrawingData()
    {
        drawingData.setLength( 0 );
    }

    /**
     * Packets are send from the client to the server.
     */
    private void activateClientPacketProcessors()
    {
        sender.addProcessor( PacketType.Client.BUTTON_ACTION, new ClientButtonActionPacketProcessor()
        {
            private static final long serialVersionUID = -5395185862321111079L;

            @Override
            public void process( ClientButtonActionPacket packet )
            {
                Window window = getWindow( packet.getWindowId() );
                if ( Objects.nonNull( window ) )
                {
                    if ( focusedWindowId != window.getId() )
                    {
                        sender.onSend( new ClientFocusPacket( window.getId() ) );
                    }
                    sender.send( packet );
                }
            }
        } );

        sender.addProcessor( PacketType.Client.CLIPBOARD_CONTENTS, new ClientClipboardContentsPacketProcessor()
        {
            private static final long serialVersionUID = -1921396659933750167L;

            @Override
            public void process( ClientClipboardContentsPacket packet )
            {
                Window window = getWindow( focusedWindowId );
                if ( Objects.nonNull( window ) && window.isClipboardEnabled() )
                {
                    sender.send( packet );
                }
            }
        } );

        sender.addProcessor( PacketType.Client.CLIPBOARD_CONTENTS_NONE, new ClientClipboardContentsNonePacketProcessor()
        {
            private static final long serialVersionUID = -7192199900873807467L;

            @Override
            public void process( ClientClipboardContentsNonePacket packet )
            {
                Window window = getWindow( focusedWindowId );
                if ( Objects.nonNull( window ) && window.isClipboardEnabled() )
                {
                    sender.send( packet );
                }
            }
        } );

        sender.addProcessor( PacketType.Client.CLIPBOARD_TOKEN, new ClientClipboardTokenPacketProcessor()
        {
            private static final long serialVersionUID = -5880468244343237489L;

            @Override
            public void process( ClientClipboardTokenPacket packet )
            {
                Window window = getWindow( focusedWindowId );
                if ( Objects.nonNull( window ) && window.isClipboardEnabled() )
                {
                    sender.send( packet );
                }
            }
        } );

        sender.addProcessor( PacketType.Client.CLOSE_WINDOW, new ClientCloseWindowPacketProcessor()
        {
            private static final long serialVersionUID = 2611012175826117454L;

            @Override
            public void process( ClientCloseWindowPacket packet )
            {
                Window window = getWindow( packet.getWindowId() );
                if ( Objects.nonNull( window ) )
                {
                    sender.send( packet );
                }
            }
        } );

        sender.addProcessor( PacketType.Client.CONFIGURE_WINDOW, new ClientConfigureWindowPacketProcessor()
        {
            private static final long serialVersionUID = -6416255007614528365L;

            @Override
            public void process( ClientConfigureWindowPacket packet )
            {
                Window window = getWindow( packet.getWindowId() );
                if ( Objects.nonNull( window ) )
                {
                    if ( window.isMaximized() )
                    {
                        window.setX( packet.getX() );
                        window.setY( packet.getHeight() );
                        window.setWidth( packet.getWidth() );
                        window.setHeight( packet.getHeight() );
                    }
                    else
                    {
                        int x = Math.max( 0, Math.min( packet.getX(), desktopWidth - packet.getWidth() ) );
                        int y = Math.max( 0, Math.min( packet.getY(), desktopHeight - packet.getHeight() ) );
                        int width = ( x + packet.getWidth() ) > desktopWidth ? ( desktopWidth - x ) : packet.getWidth();
                        int height = ( y + packet.getHeight() ) > desktopHeight ? ( desktopHeight - y ) : packet.getHeight();
                        window.setX( x );
                        window.setY( y );
                        window.setWidth( width );
                        window.setHeight( height );
                    }

                    window.setX( Math.max( 0, window.getX() ) );
                    window.setY( Math.max( 0, window.getY() ) );
                    if ( !window.isMaximized() )
                    {
                        window.setRestoredWindowDimension( window.getX(), window.getY(), window.getWidth(), window.getHeight() );
                    }
                    sender.send( packet );
                }
            }
        } );

        sender.addProcessor( PacketType.Client.DAMAGE_SEQUENCE, new ClientDamageSequencePacketProcessor()
        {
            private static final long serialVersionUID = -4025876263762950639L;

            @Override
            public void process( ClientDamageSequencePacket packet )
            {
                Window window = getWindow( packet.getWindowId() );
                if ( Objects.nonNull( window ) )
                {
                    sender.send( packet );
                }
            }
        } );

        sender.addProcessor( PacketType.Client.DESKTOP_SIZE, new ClientDesktopSizePacketProcessor()
        {

            private static final long serialVersionUID = 4556345095949490074L;

            @Override
            public void process( ClientDesktopSizePacket packet )
            {
                sender.send( packet );
            }
        } );

        sender.addProcessor( PacketType.Client.DISCONNECT, new ClientDisconnectPacketProcessor()
        {
            private static final long serialVersionUID = -95609814527210688L;

            @Override
            public void process( ClientDisconnectPacket packet )
            {
                for ( Window window : getWindows() )
                {
                    WindowDimension windowDimension = window.getRestoredWindowDimension();
                    if ( Objects.nonNull( windowDimension ) )
                    {
                        sender.send( new ClientConfigureWindowPacket( window.getId(), windowDimension.getX(), windowDimension.getY(), windowDimension.getWidth(), windowDimension.getHeight() ) );
                    }
                    else
                    {
                        sender.send( new ClientConfigureWindowPacket( window.getId(), window.getX(), window.getY(), window.getWidth(), window.getHeight() ) );
                    }
                }

                virtualClientAction( UNDEFINED_WINDOW, VirtualClientAction.CLOSE_APPLICATION );
                if ( StringUtils.isNotBlank( packet.getCause() ) )
                {
                    LOGGER.info( CONNECTION_INTERRUPTED_BY_USER, packet.getCause() );
                }
                else
                {
                    LOGGER.info( CONNECTION_INTERRUPTED );
                }
                sender.send( packet );
            }
        } );

        sender.addProcessor( PacketType.Client.FOCUS, new ClientFocusPacketProcessor()
        {
            private static final long serialVersionUID = 5439442460070645332L;

            @Override
            public void process( ClientFocusPacket packet )
            {
                if ( virtualUtil.isDrawingStarted() )
                {
                    Window window = getWindow( packet.getWindowId() );
                    if ( Objects.nonNull( window ) && window.isNotOverrideRedirect() && !WindowType.SPLASH.equals( window.getWindowType() ) && !WindowType.TOOLBAR.equals( window.getWindowType() ) )
                    {
                        focusedWindowId = window.getId();
                        createdWindowId = focusedWindowId;
                        virtualClientAction( focusedWindowId, VirtualClientAction.NEW_WINDOW_FOCUS );
                        sender.send( packet );
                    }
                }
            }
        } );

        sender.addProcessor( PacketType.Client.HELLO, new ClientHelloPacketProcessor()
        {
            private static final long serialVersionUID = 2216198733390619217L;

            @Override
            public void process( ClientHelloPacket packet )
            {
                sender.send( packet );
            }
        } );

        sender.addProcessor( PacketType.Client.KEY_ACTION, new ClientKeyActionPacketProcessor()
        {
            private static final long serialVersionUID = 4055925364207717328L;

            @Override
            public void process( ClientKeyActionPacket packet )
            {
                Window window = getWindow( packet.getWindowId() );
                if ( Objects.nonNull( window ) )
                {
                    sender.send( packet );
                }
            }
        } );

        sender.addProcessor( PacketType.Client.MAP_WINDOW, new ClientMapWindowPacketProcessor()
        {
            private static final long serialVersionUID = 6504089159491705451L;

            @Override
            public void process( ClientMapWindowPacket packet )
            {
                Window window = getWindow( packet.getWindowId() );
                if ( Objects.nonNull( window ) && window.isNotOverrideRedirect() )
                {
                    sender.send( packet );
                }
            }
        } );

        sender.addProcessor( PacketType.Client.PING, new ClientPingPacketProcessor()
        {
            private static final long serialVersionUID = -3929441094469776394L;

            @Override
            public void process( ClientPingPacket packet )
            {
                sender.send( packet );
            }
        } );

        sender.addProcessor( PacketType.Client.PING_ECHO, new ClientPingEchoPacketProcessor()
        {
            private static final long serialVersionUID = 9012271598814560845L;

            @Override
            public void process( ClientPingEchoPacket packet )
            {
                sender.send( packet );
            }
        } );

        sender.addProcessor( PacketType.Client.POINTER_POSITION, new ClientPointerPositionPacketProcessor()
        {
            private static final long serialVersionUID = 4641537905675748091L;

            @Override
            public void process( ClientPointerPositionPacket packet )
            {
                if ( Objects.nonNull( getWindow( packet.getWindowId() ) ) )
                {
                    sender.send( packet );
                }
            }
        } );

        sender.addProcessor( PacketType.Client.RESUME, new ClientResumePacketProcessor()
        {
            private static final long serialVersionUID = 4454191687162286107L;

            @Override
            public void process( ClientResumePacket packet )
            {
                List<Integer> windowIds = new ArrayList<>();
                windows.forEach( ( id, window ) -> windowIds.add( id ) );
                packet.setWindowIds( windowIds );
                sender.send( packet );
            }
        } );

        sender.addProcessor( PacketType.Client.SET_DEFLATE, new ClientSetDeflatePacketProcessor()
        {
            private static final long serialVersionUID = 5947313890137144717L;

            @Override
            public void process( ClientSetDeflatePacket packet )
            {
                sender.send( packet );
            }
        } );

        sender.addProcessor( PacketType.Client.SUSPEND, new ClientSuspendPacketProcessor()
        {
            private static final long serialVersionUID = -1838363347507697845L;

            @Override
            public void process( ClientSuspendPacket packet )
            {
                List<Integer> windowIds = new ArrayList<>();
                windows.forEach( ( id, window ) -> windowIds.add( id ) );
                packet.setWindowIds( windowIds );
                sender.send( packet );
            }
        } );

        sender.addProcessor( PacketType.Client.UNMAP_WINDOW, new ClientUnmapWindowPacketProcessor()
        {
            private static final long serialVersionUID = 397226949956922004L;

            @Override
            public void process( ClientUnmapWindowPacket packet )
            {
                Window window = getWindow( packet.getWindowId() );
                if ( Objects.nonNull( window ) )
                {
                    windows.remove( window.getId() );
                    if ( window.isNotOverrideRedirect() )
                    {
                        if ( MapUtils.isNotEmpty( windows ) )
                        {
                            if ( window.hasParent() )
                            {
                                sender.onSend( new ClientFocusPacket( window.getParentId() ) );
                            }
                            else
                            {
                                Window focusedWindow = getWindows().stream()
                                                                   .filter( w -> isNormalWindow( w ) )
                                                                   .reduce( ( firstwindow, secondWindow ) -> secondWindow )
                                                                   .orElse( windows.get( getLastWindowId() ) );
                                sender.onSend( new ClientFocusPacket( focusedWindow.getId() ) );
                            }
                        }
                        else
                        {
                            windowFocusForced = true;
                        }
                        sender.send( packet );
                    }
                }
            }
        } );

        sender.addProcessor( PacketType.Client.WHEEL_MOTION, new ClientWheelMotionPacketProcessor()
        {
            private static final long serialVersionUID = 949197326161052195L;

            @Override
            public void process( ClientWheelMotionPacket packet )
            {
                Window window = getWindow( packet.getWindowId() );
                if ( Objects.nonNull( window ) && focusedWindowId != packet.getWindowId() )
                {
                    sender.onSend( new ClientFocusPacket( window.getId() ) );
                }
                sender.send( packet );
            }
        } );
    }

    /**
     * Packets are send from the server to the client.
     */
    private void activateServerPacketProcessors()
    {
        receiver.addProcessor( PacketType.Server.CLIPBOARD_REQUEST, new ServerClipboardRequestPacketProcessor()
        {
            private static final long serialVersionUID = 1681625679330333894L;

            @Override
            public void process( ServerClipboardRequestPacket packet )
            {
                Window window = getWindow( focusedWindowId );
                if ( Objects.nonNull( window ) && window.isClipboardEnabled() && StringUtils.isNotBlank( getClipboardData() ) )
                {
                    sender.send( new ClientClipboardContentsPacket( packet.getRequestId(), packet.getSelection(), getClipboardData() ) );
                }
                else
                {
                    sender.send( new ClientClipboardContentsNonePacket( packet.getRequestId(), packet.getSelection() ) );
                }
            }
        } );

        receiver.addProcessor( PacketType.Server.CLIPBOARD_TOKEN, new ServerClipboardTokenPacketProcessor()
        {
            private static final long serialVersionUID = -4585723733687370518L;

            @Override
            public void process( ServerClipboardTokenPacket packet )
            {
                Window window = getWindow( focusedWindowId );
                if ( Objects.nonNull( window ) && Objects.nonNull( ClipboardTarget.getTarget( packet.getClipboardTarget() ) ) && window.isClipboardEnabled()
                    && Arrays.asList( ClipboardTarget.getAllTargets() )
                             .contains( ClipboardTarget.getTarget( packet.getClipboardTarget() ) )
                    && ( Objects.isNull( getClipboardData() ) || !StringUtils.equals( getClipboardData(), packet.getClipboardData() ) ) )
                {
                    setClipboardData( packet.getClipboardData() );
                    virtualClientAction( null, VirtualClientAction.CLIPBOARD_DATA, getClipboardData() );
                }
            }
        } );

        receiver.addProcessor( PacketType.Server.CONFIGURE_OVERRIDE_REDIRECT, new ServerConfigureOverrideRedirectPacketProcessor()
        {
            private static final long serialVersionUID = -4426423462002534252L;

            @Override
            public void process( ServerConfigureOverrideRedirectPacket packet )
            {
                Window window = getWindow( packet.getWindowId() );
                if ( Objects.nonNull( window ) )
                {
                    window.setX( packet.getX() );
                    window.setY( packet.getY() );
                    window.setWidth( packet.getWidth() );
                    window.setHeight( packet.getHeight() );
                    virtualClientAction( window.getId(), VirtualClientAction.MOVE_RESIZE_WINDOW );
                }
            }
        } );

        receiver.addProcessor( PacketType.Server.CURSOR, new ServerCursorPacketProcessor()
        {
            private static final long serialVersionUID = 7102078739666458546L;

            @Override
            public void process( ServerCursorPacket packet )
            {
                if ( PictureEncodingType.UNDEFINED.equals( packet.getEncodingType() ) )
                {
                    // reset all cursors
                    cursor = new Cursor( packet.getX(), packet.getY(), PictureEncodingType.UNDEFINED.toString(), PictureEncodingType.UNDEFINED.toString() );
                    virtualClientAction( -1, VirtualClientAction.DRAW_CURSOR );
                }
                else if ( packet.getX() < 0 && packet.getY() < 0 )
                {
                    // reset cursor of the active window
                    cursor = new Cursor( packet.getX(), packet.getY(), PictureEncodingType.UNDEFINED.toString(), PictureEncodingType.UNDEFINED.toString() );
                    virtualClientAction( focusedWindowId, VirtualClientAction.DRAW_CURSOR );
                }
                else if ( PictureEncodingType.JPEG.equals( packet.getEncodingType() ) || PictureEncodingType.PNG.equals( packet.getEncodingType() ) )
                {
                    // draw cursor
                    cursor = new Cursor( packet.getX(), packet.getY(), CanvasImageUtil.toBase64( packet.getImage() ), packet.getEncodingType()
                                                                                                                            .toString() );
                    virtualClientAction( 0, VirtualClientAction.DRAW_CURSOR );
                }
            }
        } );

        receiver.addProcessor( PacketType.Server.DESKTOP_SIZE, new ServerDesktopSizePacketProcessor()
        {
            private static final long serialVersionUID = 893520995575574228L;

            @Override
            public void process( ServerDesktopSizePacket packet )
            {
                // nothing to do
            }
        } );

        receiver.addProcessor( PacketType.Server.DISCONNECT, new ServerDisconnectPacketProcessor()
        {
            private static final long serialVersionUID = 951599604026962202L;

            @Override
            public void process( ServerDisconnectPacket packet )
            {
                disconnectedByClientRequest = StringUtils.equals( CONNECTION_INTERRUPTED_BY_CLIENT_REQUEST, packet.getCause() );
                disconnectedByServerDetach = StringUtils.equals( CONNECTION_INTERRUPTED_BY_SERVER_DETACH, packet.getCause() );
                disconnectedByServerError = StringUtils.equals( CONNECTION_INTERRUPTED_BY_SERVER_ERROR, packet.getCause() );
                disconnectedByServerShutdown = StringUtils.equals( CONNECTION_INTERRUPTED_BY_SERVER_SHUTDOWN, packet.getCause() );
                if ( isDisconnectedByServerShutdown() )
                {
                    virtualClientAction( UNDEFINED_WINDOW, VirtualClientAction.CLOSE_WINDOW, true );
                }
                else
                {
                    virtualClientAction( UNDEFINED_WINDOW, VirtualClientAction.CLOSE_ALL_WINDOWS );
                    windows.clear();
                }

                LOGGER.info( CONNECTION_INTERRUPTED_BY_SERVER, packet.getCause() );
                createdWindowId = UNDEFINED_WINDOW;
                disconnectedByServer = true;
            }
        } );

        receiver.addProcessor( PacketType.Server.DRAW, new ServerDrawPacketProcessor()
        {
            private static final long serialVersionUID = -1391970388404245809L;

            @Override
            public void process( ServerDrawPacket packet )
            {
                Window window = windows.get( packet.getWindowId() );
                if ( Objects.nonNull( window ) )
                {
                    if ( window.isWindowRedrawRequsted() )
                    {
                        window.setWindowRedrawRequsted( false );
                        virtualClientAction( window.getId(), VirtualClientAction.MOVE_RESIZE_WINDOW );
                    }

                    String drawData = getDrawingData();
                    if ( StringUtils.isNotBlank( drawData ) && drawingDataWindowId != window.getId() )
                    {
                        virtualClientAction( drawingDataWindowId, VirtualClientAction.DRAW_CANVAS, drawData );
                        resetDrawingData();
                    }
                    drawingDataWindowId = window.getId();

                    addDrawingData( packet );
                    if ( !packet.isFlushPacket() || packet.isLastFlushPacket() )
                    {
                        drawData = getDrawingData();
                        virtualClientAction( drawingDataWindowId, VirtualClientAction.DRAW_CANVAS, drawData );
                        resetDrawingData();
                    }
                    switch ( window.getWindowType() )
                    {
                        case DEFAULT:
                        case DIALOG:
                        case NORMAL:
                        case UTILITY:
                            createdWindowId = window.getId();
                            break;
                        default:
                            break;
                    }
                    setForceUpdateEnabled( false );
                }
                else
                {
                    LOGGER.error( CANNOT_DRAW_INTO_UNDEFINED_WINDOW );
                    sender.onSend( new ClientDamageSequencePacket( packet.getWindowId(), packet.getWidth(), packet.getHeight(), packet.getPacketSequence(),
                                                                   ClientDamageSequencePacket.SEQUENCE_FRAME_UNDEFINED, CANNOT_DRAW_INTO_UNDEFINED_WINDOW ) );
                }
            }
        } );

        receiver.addProcessor( PacketType.Server.ENCODINGS, new ServerEncodingsPacketProcessor()
        {
            private static final long serialVersionUID = -3442658775462717203L;

            @Override
            public void process( ServerEncodingsPacket packet )
            {
                // nothing to do
            }
        } );

        receiver.addProcessor( PacketType.Server.EOS, new ServerEosPacketProcessor()
        {
            private static final long serialVersionUID = -4089504271049118152L;

            @Override
            public void process( ServerEosPacket packet )
            {
                // nothing to do
            }
        } );

        receiver.addProcessor( PacketType.Server.HELLO, new ServerHelloPacketProcessor()
        {
            private static final long serialVersionUID = 7595934926283626842L;

            @Override
            public void process( ServerHelloPacket packet )
            {
                modifiers = new ArrayList<>( packet.getModifiers() );
                mousePreciseWheel = packet.getMousePreciseWheel();
                sender.setRencodeEnabled( packet.isRencodeEnabled() );
                sender.onSend( new ClientSetDeflatePacket( compressionLevel ) );
            }
        } );

        receiver.addProcessor( PacketType.Server.LOST_WINDOW, new ServerLostWindowPacketProcessor()
        {
            private static final long serialVersionUID = -8246543207564084691L;

            @Override
            public void process( ServerLostWindowPacket packet )
            {
                Window window = getWindow( packet.getWindowId() );
                if ( Objects.nonNull( window ) )
                {
                    sender.onSend( new ClientUnmapWindowPacket( window.getId() ) );
                    virtualClientAction( window.getId(), VirtualClientAction.CLOSE_WINDOW, isNormalWindow( window ) );
                }
            }
        } );

        receiver.addProcessor( PacketType.Server.NEW_OVERRIDE_REDIRECT, new ServerNewOverrideRedirectPacketProcessor()
        {
            private static final long serialVersionUID = 7148743728930242195L;

            @Override
            public void process( ServerNewOverrideRedirectPacket packet )
            {
                if ( Objects.isNull( windows.get( packet.getWindowId() ) ) )
                {
                    Window window = new Window( packet );
                    window.setModifiers( modifiers );
                    window.setMousePreciseWheel( mousePreciseWheel );
                    window.setAutoCenterEnabled( isWindowAutoCenterEnabled() );
                    window.setX( Math.max( 0, Math.min( window.getX(), desktopWidth - window.getWidth() ) ) );
                    window.setY( Math.max( 0, Math.min( window.getY(), desktopHeight - window.getHeight() ) ) );
                    windows.put( window.getId(), window );
                    sender.onSend( new ClientFocusPacket( window.getId() ) );
                    virtualClientAction( window.getId(), VirtualClientAction.CREATE_NEW_OVERRIDE_REDIRECT );

                    // set the window focus forced
                    if ( windowFocusForced && getNormalWindowsCounter() <= ONLY_ONE_WINDOW && window.getId() != focusedWindowId )
                    {
                        windowFocusForced = false;
                    }
                }
            }
        } );

        receiver.addProcessor( PacketType.Server.NEW_WINDOW, new ServerNewWindowPacketProcessor()
        {
            private static final long serialVersionUID = 7148743728930242195L;

            private int desktopX( int value )
            {
                return Math.round( 1.0f * desktopWidth / virtualUtil.getDesktopWidth() * value );
            }

            private int desktopY( int value )
            {
                return Math.round( 1.0f * desktopHeight / virtualUtil.getDesktopHeight() * value );
            }

            private int desktopWidth( int value, boolean dialogWindow )
            {
                return ( dialogWindow ? value : desktopX( value ) );
            }

            private int desktopHeight( int value, boolean dialogWindow )
            {
                return ( dialogWindow ? value : desktopY( value ) );
            }

            @Override
            public void process( ServerNewWindowPacket packet )
            {
                if ( Objects.isNull( windows.get( packet.getWindowId() ) ) )
                {
                    Window window = new Window( packet );
                    window.setWindowType( isNormalWindow( window ) && window.getParentId() > NO_PARENT_WINDOW ? WindowType.DIALOG : window.getWindowType() );
                    window.setWindowType( isDialogWindow( window ) && StringUtils.isBlank( window.getTitle() ) ? WindowType.SPLASH : window.getWindowType() );
                    window.setAutoCenterEnabled( isWindowAutoCenterEnabled() && !WindowType.UTILITY.equals( window.getWindowType() ) );
                    if ( virtualUtil.isNotApplicationTesting() )
                    {
                        boolean dialogWindow = WindowType.DIALOG.equals( window.getWindowType() );
                        String noMaximized = String.format( MAXIMIZED_NO_INDICATOR, window.getId() );
                        String yesMaximized = String.format( MAXIMIZED_YES_INDICATOR, window.getId() );
                        if ( window.isMaximized() && !StringUtils.contains( virtualUtil.getWindowsStatus(), noMaximized ) )
                        {
                            if ( StringUtils.contains( virtualUtil.getWindowsStatus(), yesMaximized ) )
                            {
                                int x = Math.max( 0, Math.min( desktopX( window.getX() ), desktopWidth - desktopWidth( window.getWidth(), dialogWindow ) ) );
                                int y = Math.max( 0, Math.min( desktopY( window.getY() ), desktopHeight - desktopHeight( window.getHeight(), dialogWindow ) ) );
                                int width = ( x + desktopWidth( window.getWidth(), dialogWindow ) ) > desktopWidth ? ( desktopWidth - x ) : desktopWidth( window.getWidth(), dialogWindow );
                                int height = ( y + desktopHeight( window.getHeight(), dialogWindow ) ) > desktopHeight ? ( desktopHeight - y ) : desktopHeight( window.getHeight(), dialogWindow );
                                window.setRestoredWindowDimension( x, y, width, height );
                            }
                            else
                            {
                                window.setRestoredWindowDimension( ( desktopWidth - ClientProperties.DEFAULT_WIDTH ) / 2, ( desktopHeight - ClientProperties.DEFAULT_HEIGHT ) / 2,
                                                                   ClientProperties.DEFAULT_WIDTH, ClientProperties.DEFAULT_HEIGHT );
                            }
                            window.setX( 0 );
                            window.setY( 0 );
                            window.setWidth( desktopWidth );
                            window.setHeight( desktopHeight );
                        }
                        else
                        {
                            int x = Math.max( 0, Math.min( desktopX( window.getX() ), desktopWidth - desktopWidth( window.getWidth(), dialogWindow ) ) );
                            int y = Math.max( 0, Math.min( desktopY( window.getY() ), desktopHeight - desktopHeight( window.getHeight(), dialogWindow ) ) );
                            int width = ( x + desktopWidth( window.getWidth(), dialogWindow ) ) > desktopWidth ? ( desktopWidth - x ) : desktopWidth( window.getWidth(), dialogWindow );
                            int height = ( y + desktopHeight( window.getHeight(), dialogWindow ) ) > desktopHeight ? ( desktopHeight - y ) : desktopHeight( window.getHeight(), dialogWindow );
                            if ( ( WindowType.DEFAULT.equals( window.getWindowType() ) || WindowType.NORMAL.equals( window.getWindowType() ) )
                                && ( window.getWidth() <= WINDOW_UNVISIBLE_WIDTH || window.getHeight() <= WINDOW_UNVISIBLE_HEIGHT ) )
                            {
                                window.setAutoCenterEnabled( false );
                                x = WINDOW_UNVISIBLE_X;
                                y = WINDOW_UNVISIBLE_Y;
                            }
                            window.setX( x );
                            window.setY( y );
                            window.setWidth( width );
                            window.setHeight( height );
                            window.setRestoredWindowDimension( x, y, width, height );
                        }
                        virtualUtil.setWindowsStatus( StringUtils.remove( virtualUtil.getWindowsStatus(), noMaximized ) );
                        virtualUtil.setWindowsStatus( StringUtils.remove( virtualUtil.getWindowsStatus(), yesMaximized ) );
                    }
                    else
                    {
                        int x = Math.max( 0, Math.min( window.getX(), desktopWidth - window.getWidth() ) );
                        int y = Math.max( 0, Math.min( window.getY(), desktopHeight - window.getHeight() ) );
                        int width = ( x + window.getWidth() ) > desktopWidth ? ( desktopWidth - x ) : window.getWidth();
                        int height = ( y + window.getHeight() ) > desktopHeight ? ( desktopHeight - y ) : window.getHeight();
                        if ( ( WindowType.DEFAULT.equals( window.getWindowType() ) || WindowType.NORMAL.equals( window.getWindowType() ) )
                            && ( window.getWidth() <= WINDOW_UNVISIBLE_WIDTH || window.getHeight() <= WINDOW_UNVISIBLE_HEIGHT ) )
                        {
                            window.setAutoCenterEnabled( false );
                            x = WINDOW_UNVISIBLE_X;
                            y = WINDOW_UNVISIBLE_Y;
                        }
                        window.setX( x );
                        window.setY( y );
                        window.setWidth( Math.max( WINDOW_MIN_WIDTH, width ) );
                        window.setHeight( Math.max( WINDOW_MIN_HEIGHT, height ) );
                    }
                    window.setClipboardEnabled( clipboardEnabled );
                    window.setModifiers( modifiers );
                    window.setMousePreciseWheel( mousePreciseWheel );
                    virtualUtil.setDrawingStarted( virtualUtil.isDrawingStarted() && isNotNormalWindow( window ) );

                    Window parentWindow = Objects.requireNonNullElse( getWindow( window.getParentId() ), window );
                    window.setDrawingStarted( isNotNormalWindow( window ) );
                    window.setVisible( parentWindow.isVisible() && isNotNormalWindow( window ) );
                    windows.put( window.getId(), window );

                    virtualClientAction( window.getId(), VirtualClientAction.CREATE_NEW_WINDOW );
                    sender.onSend( new ClientMapWindowPacket( window.getId(), window.getX(), window.getY(), window.getWidth(), window.getHeight() ) );
                    sender.onSend( new ClientConfigureWindowPacket( window.getId(), window.getX(), window.getY(), window.getWidth(), window.getHeight() ) );
                    if ( ( windowFocusForced && getNormalWindowsCounter() <= ONLY_ONE_WINDOW && window.getId() != focusedWindowId ) || ( ( isApplicationRunning() && window.getId() != focusedWindowId
                        && getNormalWindowsCounter() > ONLY_ONE_WINDOW && ( WindowType.DEFAULT.equals( window.getWindowType() ) || WindowType.NORMAL.equals( window.getWindowType() ) ) )
                        || ( getNormalWindowsCounter() < ONLY_ONE_WINDOW && WindowType.DIALOG.equals( window.getWindowType() ) ) ) )
                    {
                        windowFocusForced = false;
                        sender.onSend( new ClientFocusPacket( window.getId() ) );
                    }
                }
                else if ( !isForceUpdateEnabled() )
                {
                    LOGGER.error( WINDOW_EXISTS, packet.getWindowId() );
                }
            }
        } );

        receiver.addProcessor( PacketType.Server.PING, new ServerPingPacketProcessor()
        {
            private static final long serialVersionUID = -73401121944316394L;

            @Override
            public void process( ServerPingPacket packet )
            {
                sender.onSend( new ClientPingEchoPacket( packet ) );
            }
        } );

        receiver.addProcessor( PacketType.Server.PING_ECHO, new ServerPingEchoPacketProcessor()
        {
            private static final long serialVersionUID = 6921554889313757209L;

            @Override
            public void process( ServerPingEchoPacket packet )
            {
                // nothing to do
            }
        } );

        receiver.addProcessor( PacketType.Server.RAISE_WINDOW, new ServerRaiseWindowPacketProcessor()
        {
            private static final long serialVersionUID = -7788475521102303430L;

            @Override
            public void process( ServerRaiseWindowPacket packet )
            {
                Window window = getWindow( packet.getWindowId() );
                if ( Objects.nonNull( window ) && focusedWindowId != window.getId() )
                {
                    sender.onSend( new ClientFocusPacket( window.getId() ) );
                }
            }
        } );

        receiver.addProcessor( PacketType.Server.SET_CLIPBOARD_ENABLED, new ServerSetClipboardEnabledPacketProcessor()
        {
            private static final long serialVersionUID = 9106756125906998738L;

            @Override
            public void process( ServerSetClipboardEnabledPacket packet )
            {
                clipboardEnabled = packet.isClipboardEnabled();
            }
        } );

        receiver.addProcessor( PacketType.Server.SET_DEFLATE, new ServerSetDeflatePacketProcessor()
        {
            private static final long serialVersionUID = 5626699770394978420L;

            @Override
            public void process( ServerSetDeflatePacket packet )
            {
                compressionLevel = packet.getCompressionLevel();
            }
        } );

        receiver.addProcessor( PacketType.Server.STARTUP_COMPLETE, new ServerStartupCompletePacketProcessor()
        {
            private static final long serialVersionUID = -8748297261284122452L;

            @Override
            public void process( ServerStartupCompletePacket packet )
            {
                if ( !isForceUpdateEnabled() )
                {
                    LOGGER.info( STARTUP_COMPLETE );
                }

                // enable windows drawing
                if ( isForceUpdateEnabled() )
                {
                    virtualClientAction( UNDEFINED_WINDOW, VirtualClientAction.UPDATE_WINDOWS_FOCUS );
                }
            }
        } );

        receiver.addProcessor( PacketType.Server.WINDOW_ICON, new ServerWindowIconPacketProcessor()
        {
            private static final long serialVersionUID = -2532436813359652117L;

            @Override
            public void process( ServerWindowIconPacket packet )
            {
                Window window = getWindow( packet.getWindowId() );
                if ( Objects.nonNull( window ) )
                {
                    window.setIcon( packet.getImage() );
                    virtualClientAction( window.getId(), VirtualClientAction.DRAW_WINDOW_ICON );
                }
            }
        } );

        receiver.addProcessor( PacketType.Server.WINDOW_METADATA, new ServerWindowMetadataPacketProcessor()
        {
            private static final long serialVersionUID = 4672402332765325067L;

            @Override
            public void process( ServerWindowMetadataPacket packet )
            {
                Window window = getWindow( packet.getWindowId() );
                if ( Objects.nonNull( window ) )
                {
                    window.setTitle( packet.getTitle() );
                    window.setApplicationName( packet.getApplicationName() );
                    virtualClientAction( window.getId(), VirtualClientAction.NEW_WINDOW_TITLE );
                    if ( Objects.nonNull( packet.getIcon() ) )
                    {
                        ServerWindowIconPacket serverWindowIconPacket = packet.getIcon();
                        byte[] image = serverWindowIconPacket.getImage();
                        if ( ArrayUtils.isNotEmpty( image ) )
                        {
                            window.setIcon( image );
                            virtualClientAction( window.getId(), VirtualClientAction.DRAW_WINDOW_ICON, window.getIcon() );
                        }
                    }
                }
            }
        } );

        receiver.addProcessor( PacketType.Server.WINDOW_MOVE_RESIZE, new ServerWindowMoveResizePacketProcessor()
        {
            private static final long serialVersionUID = 5853859474342017648L;

            @Override
            public void process( ServerWindowMoveResizePacket packet )
            {
                Window window = getWindow( packet.getWindowId() );
                if ( Objects.nonNull( window ) && virtualUtil.isDrawingStarted() )
                {
                    window.setX( packet.getX() > desktopWidth ? window.getX() : packet.getX() );
                    window.setY( packet.getY() > desktopHeight ? window.getY() : packet.getY() );
                    window.setWidth( packet.getWidth() );
                    window.setHeight( packet.getHeight() );
                    virtualClientAction( window.getId(), VirtualClientAction.MOVE_RESIZE_WINDOW );
                }
            }
        } );

        receiver.addProcessor( PacketType.Server.WINDOW_RESIZED, new ServerWindowResizedPacketProcessor()
        {
            private static final long serialVersionUID = 3113287980963007443L;

            @Override
            public void process( ServerWindowResizedPacket packet )
            {
                Window window = getWindow( packet.getWindowId() );
                if ( Objects.nonNull( window ) )
                {
                    virtualClientAction( window.getId(), VirtualClientAction.RESIZE_WINDOW, window.getX(), window.getY(), packet.getWidth(), packet.getHeight(), window.isMaximized() );
                }
            }
        } );
    }

    public void onConnect( Sender sender )
    {
        this.sender = sender;
        activateClientPacketProcessors();

        ClientHelloPacket clientHelloPacket = new ClientHelloPacket( desktopWidth, desktopHeight, dpi, colorGamut, pictureEncodingTypes, rgbFormats, fullCscModes, keyboardLayout, keyCodes );
        clientHelloPacket.setDpi( dpi );
        this.sender.onSend( clientHelloPacket );
    }

    public void onDisconnect()
    {
        virtualClientAction( null, VirtualClientAction.CLOSE_ALL_WINDOWS );
        windows.clear();
        disconnectedByServer = false;
        createdWindowId = UNDEFINED_WINDOW;

        if ( Objects.nonNull( sender ) )
        {
            sender.onSend( new ClientDisconnectPacket() );
            sender = null;
        }
    }

    public boolean isApplicationRunning()
    {
        return applicationRunning;
    }

    public void setApplicationRunning( boolean applicationRunning )
    {
        this.applicationRunning = applicationRunning;
    }

    public boolean isDisconnectedBySecurityError()
    {
        return disconnectedBySecurityError;
    }

    public void setDisconnectedBySecurityError( boolean disconnectedBySecurityError )
    {
        this.disconnectedBySecurityError = disconnectedBySecurityError;
    }

    public boolean isDisconnectedByServer()
    {
        return disconnectedByServer;
    }

    public void setDisconnectedByServer( boolean disconnectedByServer )
    {
        this.disconnectedByServer = disconnectedByServer;
    }

    public boolean isDisconnectedByClientRequest()
    {
        return disconnectedByClientRequest;
    }

    public boolean isDisconnectedByServerDetach()
    {
        return disconnectedByServerDetach;
    }

    public void setDisconnectedByServerDetach( boolean disconnectedByServerDetach )
    {
        this.disconnectedByServerDetach = disconnectedByServerDetach;
    }

    public boolean isDisconnectedByServerError()
    {
        return disconnectedByServerError;
    }

    public void setDisconnectedByServerError( boolean disconnectedByServerError )
    {
        this.disconnectedByServerError = disconnectedByServerError;
    }

    public boolean isDisconnectedByServerShutdown()
    {
        return disconnectedByServerShutdown;
    }

    public boolean isForceUpdateEnabled()
    {
        return forceUpdateEnabled;
    }

    public void setForceUpdateEnabled( boolean forceUpdateEnabled )
    {
        this.forceUpdateEnabled = forceUpdateEnabled;
    }

    protected boolean isUpdateEnabled()
    {
        return updateEnabled;
    }

    public void setUpdateEnabled( boolean updateEnabled )
    {
        this.updateEnabled = updateEnabled;
    }

    public boolean isWindowAutoCenterEnabled()
    {
        return windowAutoCenterEnabled;
    }

    public void setWindowAutoCenterEnabled( boolean windowAutoCenterEnabled )
    {
        this.windowAutoCenterEnabled = windowAutoCenterEnabled;
    }

    public boolean isDialogWindow( Window window )
    {
        return ( Objects.nonNull( window ) && WindowType.DIALOG.equals( window.getWindowType() ) );
    }

    public boolean isNotDialogWindow( Window window )
    {
        return !isDialogWindow( window );
    }

    public int getDialogWindowsCounter()
    {
        return (int) getWindows().stream()
                                 .filter( this::isDialogWindow )
                                 .count();
    }

    public boolean isNormalWindow( Window window )
    {
        return ( Objects.nonNull( window ) && window.isNotOverrideRedirect() && ( WindowType.DEFAULT.equals( window.getWindowType() ) || WindowType.NORMAL.equals( window.getWindowType() ) ) );
    }

    public boolean isNotNormalWindow( Window window )
    {
        return !isNormalWindow( window );
    }

    public boolean isNormalWindowWithOverrideRedirect( Window window )
    {
        return ( Objects.nonNull( window ) && window.isOverrideRedirect() && ( WindowType.DEFAULT.equals( window.getWindowType() ) || WindowType.NORMAL.equals( window.getWindowType() ) ) );
    }

    public boolean isDrawingStartedInAllDialogOrNormalWindows()
    {
        if ( getDialogAndNormalWindowsCounter() >= ONLY_ONE_WINDOW )
        {
            return getWindows().stream()
                               .filter( window -> isDialogWindow( window ) || isNormalWindow( window ) )
                               .allMatch( Window::isDrawingStarted );
        }
        return false;
    }

    public int getSplashWindowsCounter()
    {
        return (int) getWindows().stream()
                                 .filter( this::isSplashWindow )
                                 .count();
    }

    public boolean isSplashWindow( Window window )
    {
        return ( Objects.nonNull( window ) && WindowType.SPLASH.equals( window.getWindowType() ) );
    }

    public boolean isNotSplashWindow( Window window )
    {
        return !isSplashWindow( window );
    }

    public int getDialogAndNormalWindowsCounter()
    {
        return (int) getWindows().stream()
                                 .filter( window -> isDialogWindow( window ) || isNormalWindow( window ) )
                                 .count();
    }

    public int getDialogAndNormalAndSplashWindowsCounter()
    {
        return (int) getWindows().stream()
                                 .filter( window -> isDialogWindow( window ) || isNormalWindow( window ) || isSplashWindow( window ) )
                                 .count();
    }

    public int getNormalWindowsCounter()
    {
        return (int) getWindows().stream()
                                 .filter( this::isNormalWindow )
                                 .count();
    }

    public boolean isWindowCreated()
    {
        if ( !windowCreated && createdWindowId != UNDEFINED_WINDOW )
        {
            Window window = windows.get( createdWindowId );
            if ( Objects.nonNull( window ) )
            {
                switch ( window.getWindowType() )
                {
                    case DEFAULT:
                    case DIALOG:
                    case NORMAL:
                    case UTILITY:
                        windowCreated = isUpdateEnabled();
                        break;
                    default:
                        break;
                }
            }
        }
        return windowCreated;
    }

    public boolean isNotWindowCreated()
    {
        return !isWindowCreated();
    }

    public String getClipboardData()
    {
        return clipboardData;
    }

    public void setClipboardData( String clipboardData )
    {
        this.clipboardData = clipboardData;
    }

    public void onPacketReceived( List<Object> list )
    {
        receiver.onReceive( list );
    }

    public Sender getSender()
    {
        return sender;
    }

    public Cursor getCursor()
    {
        return cursor;
    }

    public Collection<Window> getWindows()
    {
        return windows.values();
    }

    public Window getWindow( int windowId )
    {
        return windows.get( windowId );
    }

    public List<Integer> getWindowIds()
    {
        return new ArrayList<>( windows.keySet() );
    }

    public Integer getLastWindowId()
    {
        List<Integer> ids = new ArrayList<>( getWindowIds() );
        return Collections.max( ids );
    }

    public VirtualUtil getVirtualUtil()
    {
        return virtualUtil;
    }

}
