/**
 * Copyright (C) 2020-2022 Dušan Vejnović <vaadin@dussan.org>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.dussan.vaadin.dvirtual.xpra.io;

import java.io.ByteArrayOutputStream;
import java.io.Closeable;
import java.io.IOException;
import java.io.OutputStream;
import java.io.Serializable;
import java.util.ArrayList;
import java.util.Objects;

import org.dussan.vaadin.dvirtual.utils.LoggerUtil;
import org.dussan.vaadin.dvirtual.xpra.io.rencode.RencodeOutputStream;
import org.dussan.vaadin.dvirtual.xpra.packets.ClientPacket;
import org.dussan.vaadin.dvirtual.xpra.packets.PacketType;
import org.dussan.vaadin.dvirtual.xpra.processors.ClientPacketProcessors;
import org.dussan.vaadin.dvirtual.xpra.processors.base.ClientPacketProcessor;
import org.slf4j.Logger;

import com.dampcake.bencode.BencodeOutputStream;

public class Sender
    implements Closeable, Serializable
{
    private static final long serialVersionUID = -6635990390669141100L;

    private static final Logger LOGGER = LoggerUtil.getLogger();

    private static final int BYTE_ARRAY_SIZE = 8192;

    private static final String FAILED_TO_SEND_PACKAGE = "Failed to send client packet, packet type: {}.";

    private static final String PACKET_NOT_SUPPORTED = "Client packet is not supported, packet type: {}.";

    private static final String PACKET_TOO_SMALL = "Client packet is too small.";

    private boolean connectionClosed = false;

    private boolean rencodeEnabled = false;

    private ClientPacketProcessors clientPacketProcessors = null;

    private HeaderChunk headerChunk = null;

    private transient OutputStream outputStream = null;

    private Sender()
    {
        headerChunk = new HeaderChunk();
        clientPacketProcessors = new ClientPacketProcessors();
    }

    public Sender( OutputStream outputStream )
    {
        this();
        this.outputStream = outputStream;
    }

    public void addProcessor( PacketType.Client type, ClientPacketProcessor<?> processor )
    {
        clientPacketProcessors.addProcessor( type, processor );
    }

    @SuppressWarnings( "resource" )
    public synchronized void send( ClientPacket packet )
    {
        if ( !connectionClosed )
        {
            try
            {
                ByteArrayOutputStream byteArrayOutputStream = new ByteArrayOutputStream( BYTE_ARRAY_SIZE );
                BencodeOutputStream bencodeOutputStream = new BencodeOutputStream( byteArrayOutputStream );
                RencodeOutputStream rencodeOutputStream = new RencodeOutputStream( byteArrayOutputStream );

                ArrayList<Object> packetData = new ArrayList<>();
                packet.packPacket( packetData );
                if ( rencodeEnabled )
                {
                    rencodeOutputStream.writeCollection( packetData );
                    headerChunk.setFlags( HeaderChunk.FLAGS_RENCODE );
                }
                else
                {
                    bencodeOutputStream.writeList( packetData );
                    headerChunk.setFlags( 0 );
                }

                byte[] bytes = byteArrayOutputStream.toByteArray();
                int packetSize = byteArrayOutputStream.size();
                headerChunk.setPacketSize( packetSize );
                headerChunk.writeHeader( outputStream );
                outputStream.write( bytes, 0, packetSize );
                outputStream.flush();
                byteArrayOutputStream.reset();
            }
            catch ( IOException exception )
            {
                LOGGER.warn( FAILED_TO_SEND_PACKAGE, packet.getPacketType() );
            }
        }
    }

    public void setRencodeEnabled( boolean enabled )
    {
        rencodeEnabled = enabled;
    }

    @SuppressWarnings( "unchecked" )
    public synchronized void onSend( ClientPacket clientPacket )
    {
        if ( Objects.isNull( clientPacket ) )
        {
            LOGGER.info( PACKET_TOO_SMALL );
        }
        else
        {
            ClientPacketProcessor<ClientPacket> clientPacketProcessor = null;
            PacketType.Client packetType = PacketType.Client.getType( clientPacket.getPacketType() );
            switch ( packetType )
            {
                case BUTTON_ACTION:
                case CLIPBOARD_CONTENTS:
                case CLIPBOARD_CONTENTS_NONE:
                case CLIPBOARD_TOKEN:
                case CLOSE_WINDOW:
                case CONFIGURE_WINDOW:
                case DAMAGE_SEQUENCE:
                case DESKTOP_SIZE:
                case DISCONNECT:
                case FOCUS:
                case HELLO:
                case KEY_ACTION:
                case MAP_WINDOW:
                case PING:
                case PING_ECHO:
                case POINTER_POSITION:
                case RESUME:
                case SET_DEFLATE:
                case SUSPEND:
                case UNMAP_WINDOW:
                case WHEEL_MOTION:
                    clientPacketProcessor = (ClientPacketProcessor<ClientPacket>) clientPacketProcessors.getProcessor( packetType );
                    break;
                default:
                    // nothing to do
                    break;
            }

            if ( Objects.nonNull( clientPacketProcessor ) )
            {
                // LOGGER.info( "=====================================================SEND: {}", packetType );
                clientPacketProcessor.process( clientPacket );
            }
            else
            {
                LOGGER.warn( PACKET_NOT_SUPPORTED, clientPacket.getPacketType() );
            }
        }
    }

    @Override
    public void close()
    {
        connectionClosed = true;
    }

}
