/**
 * Copyright (C) 2020-2022 Dušan Vejnović <vaadin@dussan.org>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.dussan.vaadin.dvirtual.xpra.packets.client;

import java.util.Arrays;
import java.util.Collection;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.UUID;

import org.dussan.vaadin.dvirtual.xpra.info.PictureEncodingType;
import org.apache.commons.lang3.StringUtils;
import org.dussan.vaadin.dvirtual.info.ClientProperties;
import org.dussan.vaadin.dvirtual.xpra.info.ClientVersion;
import org.dussan.vaadin.dvirtual.xpra.info.FullCscMode;
import org.dussan.vaadin.dvirtual.xpra.info.RgbFormat;
import org.dussan.vaadin.dvirtual.xpra.packets.ClientPacket;
import org.dussan.vaadin.dvirtual.xpra.packets.PacketType;
import org.dussan.vaadin.dvirtual.xpra.packets.base.ClientCapabilities;
import org.dussan.vaadin.dvirtual.xpra.packets.base.ClipboardSelection;
import org.dussan.vaadin.dvirtual.xpra.packets.base.ClipboardTarget;

public class ClientHelloPacket
    extends ClientPacket
{

    private static final long serialVersionUID = -647608257284581236L;

    private static final int AUTO_REFRESH_DELAY = 500;

    private static final int COMPRESSION_LEVEL = 1;

    private static final int ENCODING_H264_MP4_SCORE_DELTA = 50;

    private static final int ENCODING_H264_SCORE_DELTA = -20;

    private static final int ENCODING_ICONS_MAX_SIZE = 40;

    private static final int ENCODING_MIN_QUALITY = 30;

    private static final int ENCODING_MPEG4_MP4_SCORE_DELTA = 50;

    private static final int ENCODING_WP8_WEBM_SCORE_DELTA = 50;

    private static final int FILE_SIZE_LIMIT = 1;

    private static final String CLIENT_TYPE = "Java";

    private static final String ENCODING_H264_MP4_YUV420P_LEVEL = "3.0";

    private static final String ENCODING_H264_MP4_YUV420P_PROFILE = "main";

    private static final String ENCODING_H264_YUV420P_LEVEL = "2.1";

    private static final String ENCODING_H264_YUV420P_PROFILE = "baseline";

    private static final String ENCODING_X264_YUV420P_PROFILE = "baseline";

    private static final String LZ4_JS_VERSION = "1.1.0";

    private static final String OS_NAME = "os.name";

    private static final String STRING_MINUS = "-";

    private transient Map<String, Object> capabilities = new LinkedHashMap<>();

    public ClientHelloPacket( int screenWidth, int screenHeight, int dpi, String colorGamut, PictureEncodingType[] encodings, RgbFormat[] formats, FullCscMode[] modes, String keyboardLayout,
                              List<List<Object>> keyCodes )
    {
        super( PacketType.Client.HELLO.toString() );
        String uuid = UUID.randomUUID()
                          .toString()
                          .replace( STRING_MINUS, StringUtils.EMPTY );

        capabilities.put( ClientCapabilities.AUTO_REFRESH_DELAY, AUTO_REFRESH_DELAY );
        capabilities.put( ClientCapabilities.BELL, false );
        capabilities.put( ClientCapabilities.BENCODE, false );
        capabilities.put( ClientCapabilities.BROTLI, false );
        capabilities.put( ClientCapabilities.CHUNKED_COMPRESSION, false );
        capabilities.put( ClientCapabilities.CLIENT_TYPE, CLIENT_TYPE );
        capabilities.put( ClientCapabilities.CLIPBOARD, true );
        capabilities.put( ClientCapabilities.CLIPBOARD_CONTENTS_SLICE_FIX, true );
        capabilities.put( ClientCapabilities.CLIPBOARD_GREEDY, true );
        capabilities.put( ClientCapabilities.CLIPBOARD_PREFERRED_TARGETS, Arrays.asList( ClipboardTarget.getAllTargets() ) );
        capabilities.put( ClientCapabilities.CLIPBOARD_SELECTIONS, Arrays.asList( ClipboardSelection.getAllSelections() ) );
        capabilities.put( ClientCapabilities.CLIPBOARD_WANT_TARGETS, true );
        capabilities.put( ClientCapabilities.COMPRESSION_LEVEL, COMPRESSION_LEVEL );
        capabilities.put( ClientCapabilities.CURSORS, true );
        capabilities.put( ClientCapabilities.DESKTOP_MODE_SIZE, Arrays.asList( screenWidth, screenHeight ) );
        capabilities.put( ClientCapabilities.DESKTOP_SIZE, Arrays.asList( screenWidth, screenHeight ) );
        capabilities.put( ClientCapabilities.DPI, dpi );
        capabilities.put( ClientCapabilities.ENCODING_COLOR_GAMUT, colorGamut );
        capabilities.put( ClientCapabilities.ENCODING_CLIENT_OPTIONS, true );
        capabilities.put( ClientCapabilities.ENCODING_EOS, false );
        capabilities.put( ClientCapabilities.ENCODING_FLUSH, true );
        capabilities.put( ClientCapabilities.ENCODING_FULL_CSC_MODES, FullCscMode.toStringMap( modes ) );
        capabilities.put( ClientCapabilities.ENCODING_GENERIC, true );
        capabilities.put( ClientCapabilities.ENCODING_H264_CABAC, false );
        capabilities.put( ClientCapabilities.ENCODING_H264_DEBLOCKING_FILTER, false );
        capabilities.put( ClientCapabilities.ENCODING_H264_FAST_DECODE, true );
        capabilities.put( ClientCapabilities.ENCODING_H264_MP4_SCORE_DELTA, ENCODING_H264_MP4_SCORE_DELTA );
        capabilities.put( ClientCapabilities.ENCODING_H264_MP4_YUV420P_LEVEL, ENCODING_H264_MP4_YUV420P_LEVEL );
        capabilities.put( ClientCapabilities.ENCODING_H264_MP4_YUV420P_PROFILE, ENCODING_H264_MP4_YUV420P_PROFILE );
        capabilities.put( ClientCapabilities.ENCODING_H264_SCORE_DELTA, ENCODING_H264_SCORE_DELTA );
        capabilities.put( ClientCapabilities.ENCODING_H264_YUV420P_LEVEL, ENCODING_H264_YUV420P_LEVEL );
        capabilities.put( ClientCapabilities.ENCODING_H264_YUV420P_PROFILE, ENCODING_H264_YUV420P_PROFILE );
        capabilities.put( ClientCapabilities.ENCODING_ICONS_MAX_SIZE, Arrays.asList( ENCODING_ICONS_MAX_SIZE, ENCODING_ICONS_MAX_SIZE ) );
        capabilities.put( ClientCapabilities.ENCODING_MPEG4_MP4_SCORE_DELTA, ENCODING_MPEG4_MP4_SCORE_DELTA );
        capabilities.put( ClientCapabilities.ENCODING_NON_SCROLL, Arrays.asList( PictureEncodingType.JPEG, PictureEncodingType.PNG, PictureEncodingType.RGB32 ) );
        capabilities.put( ClientCapabilities.ENCODING_MIN_QUALITY, ENCODING_MIN_QUALITY );
        // capabilities.put( ClientCapabilities.ENCODING_MIN_SPEED, ENCODING_MIN_SPEED );
        capabilities.put( ClientCapabilities.ENCODING_RGB24_ZLIB, false );
        capabilities.put( ClientCapabilities.ENCODING_RGB_LZ4, true );
        capabilities.put( ClientCapabilities.ENCODING_RGB_ZLIB, true );
        // capabilities.put( ClientCapabilities.ENCODING_SCROLLING, false );
        capabilities.put( ClientCapabilities.ENCODING_TRANSPARENCY, true );
        capabilities.put( ClientCapabilities.ENCODING_X264_YUV420P_PROFILE, ENCODING_X264_YUV420P_PROFILE );
        capabilities.put( ClientCapabilities.ENCODING_WP8_WEBM_SCORE_DELTA, ENCODING_WP8_WEBM_SCORE_DELTA );
        capabilities.put( ClientCapabilities.ENCODINGS, PictureEncodingType.toStringList( encodings ) );
        capabilities.put( ClientCapabilities.ENCODINGS_CORE, PictureEncodingType.toStringList( encodings ) );
        capabilities.put( ClientCapabilities.ENCODINGS_CURSOR, Arrays.asList( PictureEncodingType.PNG.toString() ) );
        capabilities.put( ClientCapabilities.ENCODINGS_PACKET, true );
        capabilities.put( ClientCapabilities.ENCODINGS_RGB_FORMATS, RgbFormat.toStringList( formats ) );
        capabilities.put( ClientCapabilities.ENCODINGS_WINDOW_ICON, Arrays.asList( PictureEncodingType.PNG.toString() ) );
        capabilities.put( ClientCapabilities.FILE_SIZE_LIMIT, FILE_SIZE_LIMIT );
        capabilities.put( ClientCapabilities.FILE_TRANSFER, false );
        capabilities.put( ClientCapabilities.KEYBOARD, true );
        capabilities.put( ClientCapabilities.LZ4, true );
        capabilities.put( ClientCapabilities.LZ4_JS_VERSION, LZ4_JS_VERSION );
        capabilities.put( ClientCapabilities.LZO, false );
        capabilities.put( ClientCapabilities.METADATA_SUPPORTED, Arrays.asList( ClientCapabilities.Metadata.FULLSCREEN, //
                                                                                ClientCapabilities.Metadata.MAXIMIZED, //
                                                                                ClientCapabilities.Metadata.ABOVE, //
                                                                                ClientCapabilities.Metadata.BELOW, //
                                                                                ClientCapabilities.Metadata.TITLE, //
                                                                                ClientCapabilities.Metadata.SIZE_HINTS, //
                                                                                ClientCapabilities.Metadata.CLASS_INSTANCE, //
                                                                                ClientCapabilities.Metadata.TRANSIENT_FOR, //
                                                                                ClientCapabilities.Metadata.WINDOW_TYPE, //
                                                                                ClientCapabilities.Metadata.HAS_ALPHA, //
                                                                                ClientCapabilities.Metadata.DECORATIONS, //
                                                                                ClientCapabilities.Metadata.OVERRIDE_DIRECT, //
                                                                                ClientCapabilities.Metadata.MODAL, //
                                                                                ClientCapabilities.Metadata.OPACITY ) );
        capabilities.put( ClientCapabilities.NAMED_CURSORS, false );
        capabilities.put( ClientCapabilities.NOTIFICATIONS, true );
        capabilities.put( ClientCapabilities.PLATFORM, System.getProperty( OS_NAME )
                                                             .toLowerCase() );
        capabilities.put( ClientCapabilities.PRINTING, false );
        capabilities.put( ClientCapabilities.RENCODE, true );
        capabilities.put( ClientCapabilities.SCREEN_SIZES, ClientProperties.getScreenSizes( screenWidth, screenHeight, dpi ) );
        capabilities.put( ClientCapabilities.SERVER_WINDOW_RESIZE, true );
        capabilities.put( ClientCapabilities.SHARE, true );
        capabilities.put( ClientCapabilities.SYSTEM_TRAY, false );
        capabilities.put( ClientCapabilities.UUID, uuid );
        capabilities.put( ClientCapabilities.WINDOW_RAISE, true );
        capabilities.put( ClientCapabilities.WINDOWS, true );
        capabilities.put( ClientCapabilities.VERSION, ClientVersion.VERSION );
        capabilities.put( ClientCapabilities.XKBMAP_KEYCODES, keyCodes );
        capabilities.put( ClientCapabilities.XKBMAP_LAYOUT, keyboardLayout );
        capabilities.put( ClientCapabilities.YAML, false );
        capabilities.put( ClientCapabilities.ZLIB, true );
    }

    public void setDpi( int dpi )
    {
        capabilities.put( ClientCapabilities.DPI, dpi );
    }

    @Override
    public void packPacket( Collection<Object> packetData )
    {
        packetData.add( getPacketType() );
        packetData.add( capabilities );
    }

}
